import numpy as np

from neuraxle.pipeline import Pipeline
from neuraxle.steps.misc import TapeCallbackFunction, \
    FitTransformCallbackStep
from neuraxle.steps.caching import JoblibValueCachingWrapper

EXPECTED_OUTPUTS = [0.0, 0.0, 0.6931471805599453, 0.6931471805599453]


class LogFitTransformCallbackStep(FitTransformCallbackStep):
    def fit_transform(self, data_inputs, expected_outputs=None):
        FitTransformCallbackStep.fit_transform(self, data_inputs, expected_outputs)
        return self, np.log(data_inputs)


def test_transform_should_use_cache(tmpdir):
    tape_transform = TapeCallbackFunction()
    tape_fit = TapeCallbackFunction()
    p = Pipeline([
        JoblibValueCachingWrapper(
            LogFitTransformCallbackStep(
                tape_transform,
                tape_fit,
                transform_function=np.log),
            tmpdir
        )
    ])

    outputs = p.transform([1, 1, 2, 2])

    assert outputs == EXPECTED_OUTPUTS
    assert tape_transform.data == [[1], [2]]
    assert tape_fit.data == []


def test_fit_transform_should_fit_then_use_cache(tmpdir):
    tape_transform = TapeCallbackFunction()
    tape_fit = TapeCallbackFunction()
    p = Pipeline([
        JoblibValueCachingWrapper(
            LogFitTransformCallbackStep(
                tape_transform,
                tape_fit,
                transform_function=np.log),
            tmpdir
        )
    ])

    p, outputs = p.fit_transform([1, 1, 2, 2], [2, 2, 4, 4])

    assert outputs == EXPECTED_OUTPUTS
    assert tape_transform.data == [[1], [2]]
    assert tape_fit.data == [([1, 1, 2, 2], [2, 2, 4, 4])]


def test_should_flush_cache_on_every_fit(tmpdir):
    tape_transform = TapeCallbackFunction()
    tape_fit = TapeCallbackFunction()
    wrapper = JoblibValueCachingWrapper(
        LogFitTransformCallbackStep(
            tape_transform,
            tape_fit,
            transform_function=np.log),
        cache_folder=tmpdir
    )
    p = Pipeline([
        wrapper
    ])
    wrapper.create_checkpoint_path()
    wrapper.write_cache(1, 10)
    wrapper.write_cache(2, 20)

    p, outputs = p.fit_transform([1, 1, 2, 2], [2, 2, 4, 4])

    assert outputs == EXPECTED_OUTPUTS
    assert tape_transform.data == [[1], [2]]
    assert tape_fit.data == [([1, 1, 2, 2], [2, 2, 4, 4])]



import numpy as np
from sklearn.metrics import mean_squared_error

from neuraxle.hyperparams.distributions import RandInt
from neuraxle.hyperparams.space import HyperparameterSpace
from neuraxle.metaopt.callbacks import MetricCallback
from neuraxle.metaopt.deprecated import RandomSearch, HyperparamsJSONRepository, AutoMLSequentialWrapper, \
    KFoldCrossValidationWrapper
from neuraxle.metaopt.random import ValidationSplitWrapper, average_kfold_scores
from neuraxle.pipeline import Pipeline
from neuraxle.steps.numpy import MultiplyByN


def test_automl_sequential_wrapper(tmpdir):
    # Setting seed for reproducibility
    np.random.seed(68)
    # Given
    data_inputs = np.array(range(100))
    expected_outputs = np.array(range(100, 200))

    hyperparameter_space = HyperparameterSpace({
        'multiplication_1__multiply_by': RandInt(1, 3),
        'multiplication_2__multiply_by': RandInt(1, 3),
        'multiplication_3__multiply_by': RandInt(1, 3),
    })

    pipeline = Pipeline([
        ('multiplication_1', MultiplyByN()),
        ('multiplication_2', MultiplyByN()),
        ('multiplication_3', MultiplyByN())
    ], cache_folder=tmpdir).set_hyperparams_space(hyperparameter_space)

    auto_ml = RandomSearch(
        KFoldCrossValidationWrapper().set_step(pipeline),
        hyperparams_repository=HyperparamsJSONRepository(tmpdir), n_iter=10
    )

    # When
    auto_ml: AutoMLSequentialWrapper = auto_ml.fit(data_inputs, expected_outputs)
    best_model: Pipeline = auto_ml.get_best_model()
    predicted_outputs = best_model.transform(data_inputs)

    # Then
    actual_mse = ((predicted_outputs - expected_outputs) ** 2).mean()
    assert actual_mse < 20000


def test_automl_sequential_wrapper_with_validation_split_wrapper(tmpdir):
    # Setting seed for reproducibility
    np.random.seed(75)
    # Given
    data_inputs = np.array(range(100))
    expected_outputs = np.array(range(100, 200))

    hyperparameter_space = HyperparameterSpace({
        'multiplication_1__multiply_by': RandInt(1, 3),
        'multiplication_2__multiply_by': RandInt(1, 3),
        'multiplication_3__multiply_by': RandInt(1, 3),
    })

    pipeline = Pipeline([
        ('multiplication_1', MultiplyByN()),
        ('multiplication_2', MultiplyByN()),
        ('multiplication_3', MultiplyByN())
    ], cache_folder=tmpdir).set_hyperparams_space(hyperparameter_space)

    random_search = RandomSearch(
        ValidationSplitWrapper(
            pipeline,
            test_size=0.2,
            scoring_function=mean_squared_error,
            run_validation_split_in_test_mode=False
        ),
        hyperparams_repository=HyperparamsJSONRepository(tmpdir),
        higher_score_is_better=False,
        n_iter=100
    )

    # When
    mse_before = ((data_inputs - expected_outputs) ** 2).mean()
    random_search: AutoMLSequentialWrapper = random_search.fit(data_inputs, expected_outputs)
    best_model: Pipeline = random_search.get_best_model()
    predicted_outputs = best_model.transform(data_inputs)

    # Then
    actual_mse = ((predicted_outputs - expected_outputs) ** 2).mean()
    assert actual_mse < mse_before



import numpy as np
import pytest
from joblib import Parallel, delayed
from neuraxle.metaopt.trial import TRIAL_STATUS, Trials
from sklearn.metrics import mean_squared_error

from neuraxle.hyperparams.distributions import Uniform, LogNormal, Normal, Choice, Quantized, LogUniform
from neuraxle.hyperparams.space import HyperparameterSpace
from neuraxle.metaopt.auto_ml import InMemoryHyperparamsRepository, AutoML, ValidationSplitter, \
    RandomSearchHyperparameterSelectionStrategy, BaseHyperparameterSelectionStrategy
from neuraxle.metaopt.callbacks import MetricCallback, ScoringCallback
from neuraxle.metaopt.tpe import TreeParzenEstimatorHyperparameterSelectionStrategy
from neuraxle.pipeline import Pipeline
from neuraxle.steps.misc import FitTransformCallbackStep
from neuraxle.steps.numpy import AddN
import os


@pytest.mark.parametrize("expected_output_mult, pipeline", [
    (3.5, Pipeline([
        FitTransformCallbackStep().set_name('callback'),
        AddN(0.).set_hyperparams_space(HyperparameterSpace({
            'add': Choice(choice_list=[0, 1.5, 2, 3.5, 4, 5, 6]),
        })),
        AddN(0.).set_hyperparams_space(HyperparameterSpace({
            'add': Choice(choice_list=[0, 1.5, 2, 3.5, 4, 5, 6]),
        }))
    ])),
    (3.5, Pipeline([
        FitTransformCallbackStep().set_name('callback'),
        AddN(0.).set_hyperparams_space(HyperparameterSpace({
            'add': Quantized(hd=Uniform(0, 10)),
        })),
        AddN(0.).set_hyperparams_space(HyperparameterSpace({
            'add': Quantized(hd=Uniform(0, 10)),
        }))
    ])),
    (3.5, Pipeline([
        FitTransformCallbackStep().set_name('callback'),
        AddN(0.).set_hyperparams_space(HyperparameterSpace({
            'add': LogUniform(min_included=1.0, max_included=5.0),
        })),
        AddN(0.).set_hyperparams_space(HyperparameterSpace({
            'add': LogUniform(min_included=1.0, max_included=5.0),
        }))
    ])),
    (3.5, Pipeline([
        FitTransformCallbackStep().set_name('callback'),
        AddN(0.).set_hyperparams_space(HyperparameterSpace({
            'add': LogNormal(log2_space_mean=1.0, log2_space_std=0.5),
        })),
        AddN(0.).set_hyperparams_space(HyperparameterSpace({
            'add': LogNormal(log2_space_mean=1.0, log2_space_std=0.5),
        }))
    ])),
    (3.5, Pipeline([
        FitTransformCallbackStep().set_name('callback'),
        AddN(0.).set_hyperparams_space(HyperparameterSpace({
            'add': Normal(mean=3.0, std=2.0, hard_clip_min=0, hard_clip_max=6),
        })),
        AddN(0.).set_hyperparams_space(HyperparameterSpace({
            'add': Normal(mean=3.0, std=2.0, hard_clip_min=0, hard_clip_max=6),
        })),
    ])),
    (3.5, Pipeline([
        FitTransformCallbackStep().set_name('callback'),
        AddN(0.).set_hyperparams_space(HyperparameterSpace({
            'add': Uniform(0, 6),
        })),
        AddN(0.).set_hyperparams_space(HyperparameterSpace({
            'add': Uniform(0, 6),
        }))
    ]))
])
def test_tpe(expected_output_mult, pipeline, tmpdir):
    # set random seed
    np.random.seed(42)

    # Given
    tpe_scores = Parallel(n_jobs=-2)(
        delayed(_test_trial_scores)(
            expected_output_mult,
            pipeline,
            TreeParzenEstimatorHyperparameterSelectionStrategy(
                number_of_initial_random_step=5,
                quantile_threshold=0.3,
                number_good_trials_max_cap=25,
                number_possible_hyperparams_candidates=100,
                prior_weight=0.,
                use_linear_forgetting_weights=False,
                number_recent_trial_at_full_weights=25
            ),
            os.path.join(tmpdir, 'tpe', str(i))
        ) for i in range(4)
    )

    random_scores = Parallel(n_jobs=-2)(
        delayed(_test_trial_scores)(
            expected_output_mult,
            pipeline,
            RandomSearchHyperparameterSelectionStrategy(),
            os.path.join(tmpdir, 'random', str(i))
        )
        for i in range(4)
    )

    mean_tpe_score = np.array(tpe_scores).flatten().mean(axis=-1)
    mean_random_score = np.array(random_scores).flatten().mean(axis=-1)

    assert mean_tpe_score < mean_random_score


def _test_trial_scores(
    expected_output_mult,
    pipeline,
    hyperparams_optimizer: BaseHyperparameterSelectionStrategy,
    tmpdir: str
):
    hp_repository: InMemoryHyperparamsRepository = InMemoryHyperparamsRepository(cache_folder=str(tmpdir))
    n_epochs = 1
    n_trials = 10
    auto_ml: AutoML = AutoML(
        pipeline=pipeline,
        hyperparams_optimizer=hyperparams_optimizer,
        validation_splitter=ValidationSplitter(0.5),
        scoring_callback=ScoringCallback(mean_squared_error, higher_score_is_better=False),
        callbacks=[
            MetricCallback('mse', metric_function=mean_squared_error, higher_score_is_better=False),
        ],
        n_trials=n_trials,
        refit_trial=True,
        epochs=n_epochs,
        hyperparams_repository=hp_repository,
        continue_loop_on_error=False
    )

    # When
    data_inputs = np.array([0, 0])
    expected_outputs = expected_output_mult * np.ones_like(data_inputs)
    auto_ml.fit(data_inputs=data_inputs, expected_outputs=expected_outputs)

    # Then
    trials: Trials = hp_repository.load_all_trials(status=TRIAL_STATUS.SUCCESS)
    validation_scores = [t.get_validation_score() for t in trials]
    return validation_scores



"""
Tests for resumable pipelines
========================================

..
    Copyright 2019, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

"""

import numpy as np
import pytest

from neuraxle.base import ExecutionContext
from neuraxle.data_container import DataContainer
from neuraxle.checkpoints import Checkpoint
from neuraxle.pipeline import Pipeline, ResumablePipeline
from neuraxle.steps.misc import TapeCallbackFunction, FitTransformCallbackStep


class SomeCheckpointStep(Checkpoint):
    def __init__(self, data_container: DataContainer = None):
        Checkpoint.__init__(self)
        self.saved = False
        self.saved_data_container = data_container
        self.checkpoint_path = None

    def set_checkpoint_path(self, path):
        self.checkpoint_path = path
        pass

    def read_checkpoint(self, data_container: DataContainer, context: ExecutionContext):
        return self.saved_data_container

    def save_checkpoint(self, data_container: DataContainer, context: ExecutionContext):
        self.saved_data_container = data_container
        self.saved = True
        return data_container

    def should_resume(self, data_container, context: ExecutionContext) -> bool:
        return self.saved_data_container is not None


class ResumablePipelineTestCase:
    def __init__(self, tape, data_inputs, expected_outputs, steps, expected_tape):
        self.steps = steps
        self.expected_outputs = expected_outputs
        self.expected_tape = expected_tape
        self.data_inputs = data_inputs
        self.tape = tape


def create_test_case(di, eo, steps, expected_tape):
    tape = TapeCallbackFunction()
    return ResumablePipelineTestCase(tape, di, eo, steps, expected_tape)


def create_test_cases():
    data_inputs = np.ones((1, 1))
    expected_outputs = np.ones((1, 1))
    dc = DataContainer(data_inputs=data_inputs, current_ids=range(len(data_inputs)), expected_outputs=expected_outputs)

    tape = TapeCallbackFunction()
    tape_fit = TapeCallbackFunction()
    tape_without_checkpoint_test_arguments = ResumablePipelineTestCase(
        tape,
        data_inputs,
        expected_outputs,
        [
            ("a", FitTransformCallbackStep(tape.callback, tape_fit.callback, ["1"])),
            ("b", FitTransformCallbackStep(tape.callback, tape_fit.callback, ["2"])),
            ("c", FitTransformCallbackStep(tape.callback, tape_fit.callback, ["3"]))
        ],
        ["1", "2", "3"])

    tape2 = TapeCallbackFunction()
    tape2_fit = TapeCallbackFunction()
    tape_checkpoint_not_saved_test_arguments = ResumablePipelineTestCase(
        tape2,
        data_inputs,
        expected_outputs,
        [
            ("a", FitTransformCallbackStep(tape2.callback, tape2_fit.callback, ["1"])),
            ("b", SomeCheckpointStep(data_container=None)
             ),
            ("c", FitTransformCallbackStep(tape2.callback, tape2_fit.callback, ["2"])),
            ("d", FitTransformCallbackStep(tape2.callback, tape2_fit.callback, ["3"]))
        ],
        ["1", "2", "3"])

    tape3 = TapeCallbackFunction()
    tape3_fit = TapeCallbackFunction()
    tape_checkpoint_saved_after_first_step_test_arguments = ResumablePipelineTestCase(
        tape3,
        data_inputs,
        expected_outputs,
        [
            ("a", FitTransformCallbackStep(tape3.callback, tape3_fit.callback, ["1"])),
            ("b", SomeCheckpointStep(data_container=dc)
             ),
            ("c", FitTransformCallbackStep(tape3.callback, tape3_fit.callback, ["2"])),
            ("d", FitTransformCallbackStep(tape3.callback, tape3_fit.callback, ["3"]))
        ],
        ["2", "3"])

    tape4 = TapeCallbackFunction()
    tape4_fit = TapeCallbackFunction()
    tape_checkpoint_saved_after_second_step_test_arguments = ResumablePipelineTestCase(
        tape4,
        data_inputs,
        expected_outputs,
        [
            ("a", FitTransformCallbackStep(tape4.callback, tape4_fit.callback, ["1"])),
            ("b", FitTransformCallbackStep(tape4.callback, tape4_fit.callback, ["2"])),
            ("c", SomeCheckpointStep(data_container=dc)
             ),
            ("d", FitTransformCallbackStep(tape4.callback, tape4_fit.callback, ["3"]))
        ],
        ["3"])

    tape5 = TapeCallbackFunction()
    tape5_fit = TapeCallbackFunction()
    tape_checkpoint_saved_after_last_step_test_arguments = ResumablePipelineTestCase(
        tape5,
        data_inputs,
        expected_outputs,
        [
            ("a", FitTransformCallbackStep(tape5.callback, tape5_fit.callback, ["1"])),
            ("b", FitTransformCallbackStep(tape5.callback, tape5_fit.callback, ["2"])),
            ("c", FitTransformCallbackStep(tape5.callback, tape5_fit.callback, ["3"])),
            ("d", SomeCheckpointStep(data_container=dc)
             ),
        ],
        [])

    tape6 = TapeCallbackFunction()
    tape6_fit = TapeCallbackFunction()
    tape_checkpoint_saved_inside_subpipeline_last_step = ResumablePipelineTestCase(
        tape6,
        data_inputs,
        expected_outputs,
        [
            ("a", FitTransformCallbackStep(tape6.callback, tape6_fit.callback, ["1"])),
            ResumablePipeline([
                ("b", FitTransformCallbackStep(tape6.callback, tape6_fit.callback, ["2"])),
                ("d", SomeCheckpointStep(data_container=dc)
                 ),
            ]),
            ("e", FitTransformCallbackStep(tape6.callback, tape6_fit.callback, ["3"])),
            ("f", FitTransformCallbackStep(tape6.callback, tape6_fit.callback, ["4"])),
        ],
        ["3", "4"])

    tape7 = TapeCallbackFunction()
    tape7_fit = TapeCallbackFunction()
    tape_checkpoint_saved_inside_subpipeline_first_step = ResumablePipelineTestCase(
        tape7,
        data_inputs,
        expected_outputs,
        [
            ("a", FitTransformCallbackStep(tape7.callback, tape7_fit.callback, ["1"])),
            ResumablePipeline([
                ("d", SomeCheckpointStep(data_container=dc)
                 ),
                ("b", FitTransformCallbackStep(tape7.callback, tape7_fit.callback, ["2"])),
            ]),
            ("e", FitTransformCallbackStep(tape7.callback, tape7_fit.callback, ["3"])),
            ("f", FitTransformCallbackStep(tape7.callback, tape7_fit.callback, ["4"])),
        ],
        ["2", "3", "4"])

    tape8 = TapeCallbackFunction()
    tape8_fit = TapeCallbackFunction()
    tape_checkpoint_saved_inside_subpipeline_step_in_the_middle = ResumablePipelineTestCase(
        tape8,
        data_inputs,
        expected_outputs,
        [
            ("a", FitTransformCallbackStep(tape8.callback, tape8_fit.callback, ["1"])),
            ResumablePipeline([
                ("b", FitTransformCallbackStep(tape8.callback, tape8_fit.callback, ["2"])),
                ("d", SomeCheckpointStep(data_container=dc)
                 ),
                ("e", FitTransformCallbackStep(tape8.callback, tape8_fit.callback, ["3"])),
            ]),
            ("f", FitTransformCallbackStep(tape8.callback, tape8_fit.callback, ["4"])),
        ],
        ["3", "4"])

    tape9 = TapeCallbackFunction()
    tape9_fit = TapeCallbackFunction()
    tape_checkpoint_saved_inside_subpipeline_of_subpipeline = ResumablePipelineTestCase(
        tape9,
        data_inputs,
        expected_outputs,
        [
            ("a", FitTransformCallbackStep(tape9.callback, tape9_fit.callback, ["1"])),
            ResumablePipeline([
                ("b", FitTransformCallbackStep(tape9.callback, tape9_fit.callback, ["2"])),
                ResumablePipeline([
                    ("e", FitTransformCallbackStep(tape9.callback, tape9_fit.callback, ["3"])),
                    ("d", SomeCheckpointStep(data_container=dc)
                     ),
                    ("f", FitTransformCallbackStep(tape9.callback, tape9_fit.callback, ["4"])),
                ]),
                ("g", FitTransformCallbackStep(tape9.callback, tape9_fit.callback, ["5"])),
            ]),
            ("h", FitTransformCallbackStep(tape9.callback, tape9_fit.callback, ["6"])),
        ],
        ["4", "5", "6"])

    tape10 = TapeCallbackFunction()
    tape10_fit = TapeCallbackFunction()
    tape_saved_checkpoint_after_another_saved_checkpoint = ResumablePipelineTestCase(
        tape10,
        data_inputs,
        expected_outputs,
        [
            ("a", FitTransformCallbackStep(tape10.callback, tape10_fit.callback, ["1"])),
            ("b", SomeCheckpointStep(data_container=dc)),
            ("c", FitTransformCallbackStep(tape10.callback, tape10_fit.callback, ["2"])),
            ("b", SomeCheckpointStep(data_container=dc)),
            ("d", FitTransformCallbackStep(tape10.callback, tape10_fit.callback, ["3"]))
        ],
        ["3"])

    tape11 = TapeCallbackFunction()
    tape11_fit = TapeCallbackFunction()
    tape_multiple_checkpoint_in_a_row = ResumablePipelineTestCase(
        tape11,
        data_inputs,
        expected_outputs,
        [
            ("a", FitTransformCallbackStep(tape11.callback, tape11_fit.callback, ["1"])),
            ("joblib_1", SomeCheckpointStep(data_container=dc)),
            ("joblib_2", SomeCheckpointStep(data_container=dc)),
            ("c", FitTransformCallbackStep(tape11.callback, tape11_fit.callback, ["2"])),
            ("d", FitTransformCallbackStep(tape11.callback, tape11_fit.callback, ["3"]))
        ],
        ["2", "3"])

    return [
        tape_without_checkpoint_test_arguments,
        tape_checkpoint_not_saved_test_arguments,
        tape_checkpoint_saved_after_first_step_test_arguments,
        tape_checkpoint_saved_after_second_step_test_arguments,
        tape_checkpoint_saved_after_last_step_test_arguments,
        tape_checkpoint_saved_inside_subpipeline_first_step,
        tape_checkpoint_saved_inside_subpipeline_last_step,
        tape_checkpoint_saved_inside_subpipeline_step_in_the_middle,
        tape_checkpoint_saved_inside_subpipeline_of_subpipeline,
        tape_saved_checkpoint_after_another_saved_checkpoint,
        tape_multiple_checkpoint_in_a_row
    ]


@pytest.mark.parametrize("test_case", create_test_cases())
def test_should_fit_transform_each_steps(test_case: ResumablePipelineTestCase, tmpdir):
    pipeline = ResumablePipeline(
        steps=test_case.steps,
        cache_folder=tmpdir
    )

    actual_pipeline, actual_data_inputs = pipeline.fit_transform(test_case.data_inputs, test_case.expected_outputs)

    actual_tape = test_case.tape.get_name_tape()
    assert isinstance(actual_pipeline, Pipeline)
    assert actual_tape == test_case.expected_tape
    assert np.array_equal(actual_data_inputs, test_case.data_inputs)


@pytest.mark.parametrize("test_case", create_test_cases())
def test_should_fit_each_steps(test_case: ResumablePipelineTestCase, tmpdir):
    pipeline = ResumablePipeline(
        steps=test_case.steps,
        cache_folder=tmpdir
    )

    actual_pipeline = pipeline.fit(test_case.data_inputs, test_case.expected_outputs)

    actual_tape = test_case.tape.get_name_tape()
    assert isinstance(actual_pipeline, Pipeline)
    assert actual_tape == test_case.expected_tape[:-1]


@pytest.mark.parametrize("test_case", create_test_cases())
def test_should_transform_each_steps(test_case: ResumablePipelineTestCase, tmpdir):
    pipeline = ResumablePipeline(
        steps=test_case.steps,
        cache_folder=tmpdir
    )

    actual_data_inputs = pipeline.transform(test_case.data_inputs)

    actual_tape = test_case.tape.get_name_tape()
    assert actual_tape == test_case.expected_tape
    assert np.array_equal(actual_data_inputs, test_case.data_inputs)



from neuraxle.base import MetaStep, BaseStep, NonTransformableMixin
from neuraxle.hyperparams.distributions import RandInt, Boolean
from neuraxle.hyperparams.space import HyperparameterSpace, HyperparameterSamples
from neuraxle.pipeline import Pipeline
from neuraxle.steps.loop import StepClonerForEachDataInput
from testing.test_pipeline import SomeStep

SOME_STEP_HP_KEY = 'somestep_hyperparam'
RAND_INT_SOME_STEP = RandInt(-10, 0)
RAND_INT_STEP_CLONER = RandInt(0, 10)

META_STEP_HP = 'metastep_hyperparam'
SOME_STEP_HP = "SomeStep__somestep_hyperparam"
META_STEP_HP_VALUE = 1
SOME_STEP_HP_VALUE = 2

HYPE_SPACE = HyperparameterSpace({
    "a__test": Boolean()
})

HYPE_SAMPLE = HyperparameterSamples({
    "a__test": True
})


class SomeMetaStep(NonTransformableMixin, MetaStep):
    pass

class SomeStepInverseTransform(SomeStep):
    def fit_transform(self, data_inputs, expected_outputs=None):
        return self, 'fit_transform'

    def inverse_transform(self, processed_outputs):
        return 'inverse_transform'


def test_step_cloner_should_get_hyperparams():
    p = StepClonerForEachDataInput(SomeStep())
    p.set_hyperparams({
        META_STEP_HP: META_STEP_HP_VALUE,
        SOME_STEP_HP: SOME_STEP_HP_VALUE
    })

    hyperparams = p.get_hyperparams()

    assert hyperparams[META_STEP_HP] == META_STEP_HP_VALUE
    assert hyperparams[SOME_STEP_HP] == SOME_STEP_HP_VALUE


def test_step_cloner_should_set_hyperparams():
    p = StepClonerForEachDataInput(SomeStep())

    p.set_hyperparams({
        META_STEP_HP: META_STEP_HP_VALUE,
        SOME_STEP_HP: SOME_STEP_HP_VALUE
    })

    assert isinstance(p.hyperparams, HyperparameterSamples)
    assert p.hyperparams[META_STEP_HP] == META_STEP_HP_VALUE
    assert p.get_step().get_hyperparams()[SOME_STEP_HP_KEY] == SOME_STEP_HP_VALUE


def test_step_cloner_update_hyperparams_should_update_step_cloner_hyperparams():
    p = StepClonerForEachDataInput(SomeStep())
    p.set_hyperparams({
        META_STEP_HP: META_STEP_HP_VALUE,
        SOME_STEP_HP: SOME_STEP_HP_VALUE
    })

    p.update_hyperparams({
        META_STEP_HP: META_STEP_HP_VALUE + 1,
    })

    assert isinstance(p.hyperparams, HyperparameterSamples)
    assert p.hyperparams[META_STEP_HP] == META_STEP_HP_VALUE + 1
    assert p.get_step().get_hyperparams()[SOME_STEP_HP_KEY] == SOME_STEP_HP_VALUE


def test_step_cloner_update_hyperparams_should_update_wrapped_step_hyperparams():
    p = StepClonerForEachDataInput(SomeStep())
    p.set_hyperparams({
        META_STEP_HP: META_STEP_HP_VALUE,
        SOME_STEP_HP: SOME_STEP_HP_VALUE
    })

    p.update_hyperparams({
        SOME_STEP_HP: SOME_STEP_HP_VALUE + 1
    })

    assert isinstance(p.hyperparams, HyperparameterSamples)
    assert p.hyperparams[META_STEP_HP] == META_STEP_HP_VALUE
    assert p.get_step().get_hyperparams()[SOME_STEP_HP_KEY] == SOME_STEP_HP_VALUE + 1


def test_step_cloner_update_hyperparams_space_should_update_step_cloner_hyperparams():
    p = StepClonerForEachDataInput(SomeStep())

    p.set_hyperparams_space({
        META_STEP_HP: RAND_INT_META_STEP,
        SOME_STEP_HP: RAND_INT_SOME_STEP
    })

    update_meta_step_hp_space = RandInt(0, 40)
    p.update_hyperparams_space({
        META_STEP_HP: update_meta_step_hp_space
    })

    assert isinstance(p.hyperparams_space, HyperparameterSpace)
    assert p.hyperparams_space[META_STEP_HP] == update_meta_step_hp_space
    assert p.wrapped.get_hyperparams_space()[SOME_STEP_HP_KEY] == RAND_INT_SOME_STEP


def test_step_cloner_update_hyperparams_space_should_update_wrapped_step_hyperparams():
    p = StepClonerForEachDataInput(SomeStep())
    p.set_hyperparams_space(HyperparameterSpace({
        META_STEP_HP: RAND_INT_META_STEP,
        SOME_STEP_HP: RAND_INT_SOME_STEP
    }))

    updated_some_step_hp_space = RandInt(0, 400)
    p.update_hyperparams_space({
        SOME_STEP_HP: updated_some_step_hp_space
    })

    assert isinstance(p.hyperparams, HyperparameterSamples)
    assert p.hyperparams_space[META_STEP_HP] == RAND_INT_META_STEP
    assert p.wrapped.get_hyperparams_space()[SOME_STEP_HP_KEY] == updated_some_step_hp_space


def test_step_cloner_should_set_steps_hyperparams():
    p = StepClonerForEachDataInput(SomeStep())

    p.set_hyperparams({
        META_STEP_HP: META_STEP_HP_VALUE,
        SOME_STEP_HP: SOME_STEP_HP_VALUE
    })

    assert isinstance(p.hyperparams, HyperparameterSamples)
    assert isinstance(p.get_step().hyperparams, HyperparameterSamples)
    assert p.get_step().get_hyperparams()[SOME_STEP_HP_KEY] == SOME_STEP_HP_VALUE


def test_step_cloner_should_set_steps_hyperparams_space():
    p = StepClonerForEachDataInput(SomeStep())

    p.set_hyperparams_space({
        META_STEP_HP: RAND_INT_STEP_CLONER,
        SOME_STEP_HP: RAND_INT_SOME_STEP
    })

    assert isinstance(p.get_step().hyperparams_space, HyperparameterSpace)
    assert p.get_step().hyperparams_space[SOME_STEP_HP_KEY] == RAND_INT_SOME_STEP


def test_step_cloner_should_set_hyperparams_space():
    p = StepClonerForEachDataInput(SomeStep())

    p.set_hyperparams_space({
        META_STEP_HP: RAND_INT_STEP_CLONER,
        SOME_STEP_HP: RAND_INT_SOME_STEP
    })

    assert isinstance(p.hyperparams_space, HyperparameterSpace)
    assert p.hyperparams_space[META_STEP_HP] == RAND_INT_STEP_CLONER
    assert p.get_step().hyperparams_space[SOME_STEP_HP_KEY] == RAND_INT_SOME_STEP


def test_step_cloner_should_get_hyperparams_space():
    p = StepClonerForEachDataInput(SomeStep())
    p.set_hyperparams_space(HyperparameterSpace({
        META_STEP_HP: RAND_INT_STEP_CLONER,
        SOME_STEP_HP: RAND_INT_SOME_STEP
    }))

    hyperparams_space = p.get_hyperparams_space()

    assert hyperparams_space[META_STEP_HP] == RAND_INT_STEP_CLONER
    assert hyperparams_space[SOME_STEP_HP] == RAND_INT_SOME_STEP


RAND_INT_META_STEP = RandInt(0, 10)


def test_pipeline_should_set_hyperparams():
    p = Pipeline([
        SomeStep().set_name('step_1'),
        SomeStep().set_name('step_2')
    ])

    p.set_hyperparams({
        'hp': 1,
        'step_1__hp': 2,
        'step_2__hp': 3
    })

    assert isinstance(p.hyperparams, HyperparameterSamples)
    assert p.hyperparams['hp'] == 1
    assert p[0].hyperparams['hp'] == 2
    assert p[1].hyperparams['hp'] == 3


def test_pipeline_should_get_hyperparams():
    p = Pipeline([
        SomeStep().set_name('step_1'),
        SomeStep().set_name('step_2')
    ])
    p.set_hyperparams({
        'hp': 1,
        'step_1__hp': 2,
        'step_2__hp': 3
    })

    hyperparams = p.get_hyperparams()

    assert isinstance(hyperparams, HyperparameterSamples)
    assert hyperparams['hp'] == 1
    assert hyperparams['step_1__hp'] == 2
    assert hyperparams['step_2__hp'] == 3


def test_pipeline_should_get_hyperparams_space():
    p = Pipeline([
        SomeStep().set_name('step_1'),
        SomeStep().set_name('step_2')
    ])
    p.set_hyperparams_space({
        'hp': RandInt(1, 2),
        'step_1__hp': RandInt(2, 3),
        'step_2__hp': RandInt(3, 4)
    })

    hyperparams_space = p.get_hyperparams_space()

    assert isinstance(hyperparams_space, HyperparameterSpace)

    assert hyperparams_space['hp'].min_included == 1
    assert hyperparams_space['hp'].max_included == 2

    assert hyperparams_space['step_1__hp'].min_included == 2
    assert hyperparams_space['step_1__hp'].max_included == 3

    assert hyperparams_space['step_2__hp'].min_included == 3
    assert hyperparams_space['step_2__hp'].max_included == 4


def test_pipeline_should_update_hyperparams():
    p = Pipeline([
        SomeStep().set_name('step_1'),
        SomeStep().set_name('step_2')
    ])

    p.set_hyperparams({
        'hp': 1,
        'step_1__hp': 2,
        'step_2__hp': 3
    })

    p.update_hyperparams({
        'hp': 4,
        'step_2__hp': 6
    })

    assert isinstance(p.hyperparams, HyperparameterSamples)
    assert p.hyperparams['hp'] == 4
    assert p[0].hyperparams['hp'] == 2
    assert p[1].hyperparams['hp'] == 6


def test_pipeline_should_update_hyperparams_space():
    p = Pipeline([
        SomeStep().set_name('step_1'),
        SomeStep().set_name('step_2')
    ])

    p.set_hyperparams_space({
        'hp': RandInt(1, 2),
        'step_1__hp': RandInt(2, 3),
        'step_2__hp': RandInt(3, 4)
    })
    p.update_hyperparams_space({
        'hp': RandInt(4, 6),
        'step_2__hp': RandInt(6, 8)
    })

    assert isinstance(p.hyperparams_space, HyperparameterSpace)

    assert p.hyperparams_space['hp'].min_included == 4
    assert p.hyperparams_space['hp'].max_included == 6

    assert p[0].hyperparams_space['hp'].min_included == 2
    assert p[0].hyperparams_space['hp'].max_included == 3

    assert p[1].hyperparams_space['hp'].min_included == 6
    assert p[1].hyperparams_space['hp'].max_included == 8


def test_meta_step_mixin_should_get_hyperparams():
    p = SomeMetaStep(SomeStep())
    p.set_hyperparams(HyperparameterSamples({
        META_STEP_HP: META_STEP_HP_VALUE,
        SOME_STEP_HP: SOME_STEP_HP_VALUE
    }))

    hyperparams = p.get_hyperparams()

    assert hyperparams[META_STEP_HP] == META_STEP_HP_VALUE
    assert hyperparams[SOME_STEP_HP] == SOME_STEP_HP_VALUE


def test_meta_step_mixin_should_set_hyperparams():
    p = SomeMetaStep(SomeStep())

    p.set_hyperparams({
        META_STEP_HP: META_STEP_HP_VALUE,
        SOME_STEP_HP: SOME_STEP_HP_VALUE
    })

    assert isinstance(p.hyperparams, HyperparameterSamples)
    assert p.hyperparams[META_STEP_HP] == META_STEP_HP_VALUE
    assert p.get_step().get_hyperparams()['somestep_hyperparam'] == SOME_STEP_HP_VALUE


def test_meta_step_mixin_update_hyperparams_should_update_meta_step_hyperparams():
    p = SomeMetaStep(SomeStep())
    p.set_hyperparams(HyperparameterSamples({
        META_STEP_HP: META_STEP_HP_VALUE,
        SOME_STEP_HP: SOME_STEP_HP_VALUE
    }))

    p.update_hyperparams({
        META_STEP_HP: META_STEP_HP_VALUE + 1
    })

    assert p.hyperparams[META_STEP_HP] == META_STEP_HP_VALUE + 1
    assert p.get_step().get_hyperparams()['somestep_hyperparam'] == SOME_STEP_HP_VALUE


def test_meta_step_mixin_update_hyperparams_should_update_wrapped_step_hyperparams():
    p = SomeMetaStep(SomeStep())
    p.set_hyperparams(HyperparameterSamples({
        META_STEP_HP: META_STEP_HP_VALUE,
        SOME_STEP_HP: SOME_STEP_HP_VALUE
    }))

    p.update_hyperparams({
        SOME_STEP_HP: SOME_STEP_HP_VALUE + 1
    })

    assert p.hyperparams[META_STEP_HP] == META_STEP_HP_VALUE
    assert p.get_step().get_hyperparams()['somestep_hyperparam'] == SOME_STEP_HP_VALUE + 1


def test_meta_step_mixin_update_hyperparams_space_should_update_meta_step_hyperparams():
    p = SomeMetaStep(SomeStep())
    p.set_hyperparams_space(HyperparameterSpace({
        META_STEP_HP: RAND_INT_META_STEP,
        SOME_STEP_HP: RAND_INT_SOME_STEP
    }))

    update_meta_step_hp_space = RandInt(0, 100)
    p.update_hyperparams_space({
        META_STEP_HP: update_meta_step_hp_space
    })

    assert p.hyperparams_space[META_STEP_HP] == update_meta_step_hp_space
    assert p.wrapped.get_hyperparams_space()['somestep_hyperparam'] == RAND_INT_SOME_STEP


def test_meta_step_mixin_update_hyperparams_space_should_update_wrapped_step_hyperparams():
    p = SomeMetaStep(SomeStep())
    p.set_hyperparams_space({
        META_STEP_HP: RAND_INT_META_STEP,
        SOME_STEP_HP: RAND_INT_SOME_STEP
    })

    updated_some_step_hp_space = RandInt(0, 100)
    p.update_hyperparams_space({
        SOME_STEP_HP: updated_some_step_hp_space
    })

    assert p.hyperparams_space[META_STEP_HP] == RAND_INT_META_STEP
    assert p.wrapped.get_hyperparams_space()['somestep_hyperparam'] == updated_some_step_hp_space


def test_meta_step_mixin_should_set_hyperparams_space():
    p = SomeMetaStep(SomeStep())
    p.set_hyperparams_space({
        META_STEP_HP: RAND_INT_META_STEP,
        SOME_STEP_HP: RAND_INT_SOME_STEP
    })

    assert p.hyperparams_space[META_STEP_HP] == RAND_INT_META_STEP
    assert p.get_step().hyperparams_space[SOME_STEP_HP_KEY] == RAND_INT_SOME_STEP


def test_meta_step_mixin_should_get_hyperparams_space():
    p = SomeMetaStep(SomeStep())
    p.set_hyperparams_space(HyperparameterSpace({
        META_STEP_HP: RAND_INT_META_STEP,
        SOME_STEP_HP: RAND_INT_SOME_STEP
    }))

    hyperparams_space = p.get_hyperparams_space()

    assert hyperparams_space[META_STEP_HP] == RAND_INT_META_STEP
    assert hyperparams_space[SOME_STEP_HP] == RAND_INT_SOME_STEP


def test_get_set_params_base_step():
    s = SomeStep()

    s.set_params(learning_rate=0.1)
    hyperparams = s.get_params()

    assert hyperparams == {"learning_rate": 0.1}



import numpy as np
import pytest

from neuraxle.base import ExecutionContext, DEFAULT_CACHE_FOLDER
from neuraxle.data_container import DataContainer
from neuraxle.metaopt.deprecated import RandomSearch
from neuraxle.metaopt.random import ValidationSplitWrapper
from neuraxle.steps.misc import FitTransformCallbackStep, TapeCallbackFunction


@pytest.mark.skip("There seems to be an error with the reference to the callback function")
def test_validation_split_wrapper_with_random_search(tmpdir):
    transform_callback = TapeCallbackFunction()
    fit_callback = TapeCallbackFunction()
    random_search = RandomSearch(
        FitTransformCallbackStep(
            transform_callback_function=transform_callback,
            fit_callback_function=fit_callback,
            transform_function=lambda di: di * 2
        ),
        ValidationSplitWrapper(test_size=0.1)
    )
    data_inputs = np.random.randint(1, 100, (100, 5))
    expected_outputs = np.random.randint(1, 100, (100, 5))

    random_search, outputs = random_search.fit_transform(data_inputs, expected_outputs)

    assert np.array_equal(outputs, data_inputs * 2)

    # should fit on train split
    assert np.array_equal(fit_callback.data[0][0], data_inputs[0:90])
    assert np.array_equal(fit_callback.data[0][1], expected_outputs[0:90])

    # should transform on test split
    assert np.array_equal(transform_callback.data[0], data_inputs[0:90])
    assert np.array_equal(transform_callback.data[1], data_inputs[90:])

    # should transform on all data at the end
    assert np.array_equal(transform_callback.data[2], data_inputs)

    assert random_search.best_model.scores_train is not None
    assert random_search.best_model.scores_validation is not None
    assert random_search.best_model.scores_train_mean is not None
    assert random_search.best_model.scores_validation_mean is not None
    assert random_search.best_model.scores_train_std is not None
    assert random_search.best_model.scores_validation_std is not None


def test_validation_split_wrapper_handle_methods_should_split_data():
    transform_callback = TapeCallbackFunction()
    fit_callback = TapeCallbackFunction()
    validation_split_wrapper = ValidationSplitWrapper(
        FitTransformCallbackStep(
            transform_callback_function=transform_callback,
            fit_callback_function=fit_callback,
            transform_function=lambda di: di * 2
        ),
        test_size=0.1
    )
    data_inputs = np.random.randint(1, 100, (100, 5))
    expected_outputs = np.random.randint(1, 100, (100, 5))

    validation_split_wrapper, outputs = validation_split_wrapper.handle_fit_transform(
        DataContainer(data_inputs=data_inputs, current_ids=list(range(len(data_inputs))),
                      expected_outputs=expected_outputs),
        ExecutionContext(DEFAULT_CACHE_FOLDER)
    )

    assert np.array_equal(outputs.data_inputs, data_inputs * 2)

    # should fit on train split
    assert np.array_equal(fit_callback.data[0][0], data_inputs[0:90])
    assert np.array_equal(fit_callback.data[0][1], expected_outputs[0:90])

    # should transform on test split
    assert np.array_equal(transform_callback.data[0], data_inputs[0:90])
    assert np.array_equal(transform_callback.data[1], data_inputs[90:])

    # should transform on all data at the end
    assert np.array_equal(transform_callback.data[2], data_inputs)

    assert validation_split_wrapper.scores_train is not None
    assert validation_split_wrapper.scores_validation is not None
    assert validation_split_wrapper.scores_train_mean is not None
    assert validation_split_wrapper.scores_validation_mean is not None
    assert validation_split_wrapper.scores_train_std is not None
    assert validation_split_wrapper.scores_validation_std is not None



import os
import shutil
from abc import ABC, abstractmethod

import numpy as np
import pytest
from sklearn.metrics import mean_squared_error

from neuraxle.base import Identity, ExecutionContext, ForceHandleMixin, StepWithContext, ForceHandleIdentity
from neuraxle.data_container import DataContainer
from neuraxle.metaopt.auto_ml import InMemoryHyperparamsRepository, AutoML, RandomSearchHyperparameterSelectionStrategy, \
    ValidationSplitter, HyperparamsJSONRepository
from neuraxle.metaopt.callbacks import ScoringCallback
from neuraxle.pipeline import Pipeline


class SomeBaseService(ABC):
    @abstractmethod
    def service_method(self, data):
        pass


class SomeService(SomeBaseService):
    def service_method(self, data):
        self.data = data


class SomeStep(ForceHandleIdentity):
    def _will_process(self, data_container: DataContainer, context: ExecutionContext):
        data_container, context = super()._will_process(data_container, context)
        service = context.get_service(SomeBaseService)
        service.service_method(data_container.data_inputs)
        return data_container, context


class RegisterServiceDynamically(ForceHandleIdentity):
    def _will_process(self, data_container: DataContainer, context: ExecutionContext):
        context.register_service(SomeBaseService, SomeService())
        return data_container, context


def test_with_context_should_inject_dependencies_properly(tmpdir):
    data_inputs = np.array([0, 1, 2, 3])
    context = ExecutionContext(root=tmpdir)
    service = SomeService()
    context.set_service_locator({SomeBaseService: service})
    p = Pipeline([
        SomeStep().assert_has_services(SomeBaseService),
        SomeStep().assert_has_services_at_execution(SomeBaseService)
    ]).with_context(context=context)

    p.transform(data_inputs=data_inputs)

    assert np.array_equal(service.data, data_inputs)


def test_with_context_should_fail_at_init_when_services_are_missing(tmpdir):
    context = ExecutionContext(root=tmpdir)
    p = Pipeline([
        SomeStep().assert_has_services(SomeBaseService)
    ]).with_context(context=context)

    data_inputs = np.array([0, 1, 2, 3])
    with pytest.raises(AssertionError) as exception_info:
        p.transform(data_inputs=data_inputs)

    assert 'SomeBaseService dependency missing' in exception_info.value.args[0]


def test_localassert_should_assert_dependencies_properly_at_exec(tmpdir):
    data_inputs = np.array([0, 1, 2, 3])
    context = ExecutionContext(root=tmpdir)
    p = Pipeline([
        RegisterServiceDynamically(),
        SomeStep().assert_has_services_at_execution(SomeBaseService)
    ]).with_context(context=context)

    p.transform(data_inputs=data_inputs)
    service = context.get_service(SomeBaseService)
    assert np.array_equal(service.data, data_inputs)


def test_localassert_should_fail_when_services_are_missing_at_exec(tmpdir):
    context = ExecutionContext(root=tmpdir)
    p = Pipeline([
        SomeStep().assert_has_services_at_execution(SomeBaseService)
    ]).with_context(context=context)
    data_inputs = np.array([0, 1, 2, 3])

    with pytest.raises(AssertionError) as exception_info:
        p.transform(data_inputs=data_inputs)

    assert 'SomeBaseService dependency missing' in exception_info.value.args[0]


def _make_autoML_loop(tmpdir, p: Pipeline):
    hp_repository = HyperparamsJSONRepository(cache_folder=tmpdir)
#    hp_repository = InMemoryHyperparamsRepository(cache_folder=str(tmpdir) + "_hp")
    n_epochs = 1
    return AutoML(
        pipeline=p,
        hyperparams_optimizer=RandomSearchHyperparameterSelectionStrategy(),
        validation_splitter=ValidationSplitter(0.20),
        scoring_callback=ScoringCallback(mean_squared_error, higher_score_is_better=False),
        n_trials=1,
        refit_trial=True,
        epochs=n_epochs,
        hyperparams_repository=hp_repository,
        cache_folder_when_no_handle=str(tmpdir),
        continue_loop_on_error=False
    )

class TestServiceAssertion:

    def _setup(self, tmpdir):
        self.tmpdir = str(tmpdir)
        self.tmpdir_hp = self.tmpdir+"_hp"

    def teardown(self):
        shutil.rmtree(self.tmpdir, ignore_errors=True)
        shutil.rmtree(self.tmpdir+"_hp", ignore_errors=True)

    def test_step_with_context_saver(self, tmpdir):
        self._setup(tmpdir)
        context = ExecutionContext(root=self.tmpdir)
        service = SomeService()
        pipeline_name = 'testname'
        context.set_service_locator({SomeBaseService: service})
        p = Pipeline([
            SomeStep().assert_has_services(SomeBaseService)
        ]).with_context(context=context)
        p.set_name(pipeline_name)
        p.save(context, full_dump=True)

        p: StepWithContext = ExecutionContext(root=self.tmpdir).load(pipeline_name)
        assert isinstance(p, StepWithContext)

        p: Pipeline = ExecutionContext(root=self.tmpdir).load(os.path.join(pipeline_name, 'Pipeline'))
        assert isinstance(p, Pipeline)


    def test_auto_ml_should_inject_dependencies_properly(self,tmpdir):
        self._setup(tmpdir)
        data_inputs = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
        expected_outputs = data_inputs * 2
        p = Pipeline([
            SomeStep().assert_has_services(SomeBaseService),
            SomeStep().assert_has_services_at_execution(SomeBaseService)
        ])
        context = ExecutionContext(root=self.tmpdir)
        service = SomeService()
        context.set_service_locator({SomeBaseService: service})

        auto_ml: AutoML = _make_autoML_loop(self.tmpdir_hp, p)
        auto_ml: StepWithContext = auto_ml.with_context(context=context)
        assert isinstance(auto_ml, StepWithContext)
        auto_ml.fit(data_inputs, expected_outputs)

        assert np.array_equal(service.data, data_inputs)


    def test_auto_ml_should_fail_at_init_when_services_are_missing(self, tmpdir):
        self._setup(tmpdir)
        data_inputs = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
        expected_outputs = data_inputs * 2
        p = Pipeline([
            RegisterServiceDynamically(),
            SomeStep().assert_has_services(SomeBaseService),
        ])

        context = ExecutionContext(root=self.tmpdir)

        auto_ml: AutoML = _make_autoML_loop(self.tmpdir_hp, p)
        auto_ml: StepWithContext = auto_ml.with_context(context=context)
        assert isinstance(auto_ml, StepWithContext)

        with pytest.raises(AssertionError) as exception_info:
            auto_ml.fit(data_inputs, expected_outputs)

        assert 'SomeBaseService dependency missing' in exception_info.value.args[0]


    def test_auto_ml_should_fail_at_exec_when_services_are_missing(self, tmpdir):
        self._setup(tmpdir)
        data_inputs = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
        expected_outputs = data_inputs * 2
        p = Pipeline([
            SomeStep().assert_has_services_at_execution(SomeBaseService),
        ])
        context = ExecutionContext(root=self.tmpdir)

        auto_ml: AutoML = _make_autoML_loop(self.tmpdir_hp, p)
        auto_ml: StepWithContext = auto_ml.with_context(context=context)
        assert isinstance(auto_ml, StepWithContext)

        with pytest.raises(AssertionError) as exception_info:
            auto_ml.fit(data_inputs, expected_outputs)
        assert 'SomeBaseService dependency missing' in exception_info.value.args[0]


    def test_auto_ml_should_assert_dependecies_properly_at_exec(self, tmpdir):
        self._setup(tmpdir)
        data_inputs = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
        expected_outputs = data_inputs * 2
        p = Pipeline([
            RegisterServiceDynamically(),
            SomeStep().assert_has_services_at_execution(SomeBaseService),
        ])
        context = ExecutionContext(root=self.tmpdir)

        auto_ml: AutoML = _make_autoML_loop(self.tmpdir_hp, p)
        auto_ml: StepWithContext = auto_ml.with_context(context=context)
        assert isinstance(auto_ml, StepWithContext)
        auto_ml.fit(data_inputs, expected_outputs)

        service = context.get_service(SomeBaseService)
        assert np.array_equal(service.data, data_inputs)



import os

import numpy as np
from joblib import dump
from py._path.local import LocalPath

from neuraxle.base import BaseStep, TruncableJoblibStepSaver
from neuraxle.checkpoints import DefaultCheckpoint
from neuraxle.hyperparams.space import HyperparameterSamples
from neuraxle.pipeline import ResumablePipeline
from neuraxle.steps.numpy import MultiplyByN

OUTPUT = "OUTPUT"
ROOT = 'ResumablePipeline'
CHECKPOINT = 'checkpoint'
SOME_STEP_2 = 'some_step2'
PIPELINE_2 = 'pipeline2'
SOME_STEP_1 = 'some_step1'
SOME_STEP_3 = 'some_step3'

EXPECTED_OUTPUTS = [0, 48, 96, 144, 192, 240, 288, 336, 384, 432]


def create_some_step3_path(tmpdir, create_dir=False):
    p = os.path.join(tmpdir, ROOT, PIPELINE_2, SOME_STEP_3, '{0}.joblib'.format(SOME_STEP_3))
    if create_dir:
        os.makedirs(os.path.join(tmpdir, ROOT, PIPELINE_2, SOME_STEP_3))
    return p


def create_some_step2_path(tmpdir, create_dir=False):
    p = os.path.join(tmpdir, ROOT, PIPELINE_2, SOME_STEP_2, '{0}.joblib'.format(SOME_STEP_2))
    if create_dir:
        os.makedirs(os.path.join(tmpdir, ROOT, PIPELINE_2, SOME_STEP_2))
    return p

def create_some_checkpoint_path(tmpdir, create_dir=False):
    p = os.path.join(tmpdir, ROOT, PIPELINE_2, CHECKPOINT, '{0}.joblib'.format(CHECKPOINT))
    if create_dir:
        os.makedirs(os.path.join(tmpdir, ROOT, PIPELINE_2, CHECKPOINT))
    return p


def create_pipeline2_path(tmpdir, create_dir=False):
    p = os.path.join(tmpdir, ROOT, PIPELINE_2, '{0}.joblib'.format(PIPELINE_2))
    if create_dir:
        os.makedirs(os.path.join(tmpdir, ROOT, PIPELINE_2))
    return p


def create_some_step1_path(tmpdir, create_dir=False):
    p = os.path.join(tmpdir, ROOT, SOME_STEP_1, '{0}.joblib'.format(SOME_STEP_1))
    if create_dir:
        os.makedirs(os.path.join(tmpdir, ROOT, SOME_STEP_1))
    return p


def create_root_path(tmpdir, create_dir=False):
    p = os.path.join(tmpdir, ROOT, '{0}.joblib'.format(ROOT))
    if create_dir and not os.path.exists(os.path.join(tmpdir, ROOT)):
        os.makedirs(os.path.join(tmpdir, ROOT))
    return p


def test_resumable_pipeline_fit_transform_should_save_all_fitted_pipeline_steps(tmpdir: LocalPath):
    p = ResumablePipeline([
        (SOME_STEP_1, MultiplyByN(multiply_by=2)),
        (PIPELINE_2, ResumablePipeline([
            (SOME_STEP_2, MultiplyByN(multiply_by=4)),
            (CHECKPOINT, DefaultCheckpoint()),
            (SOME_STEP_3, MultiplyByN(multiply_by=6))
        ]))
    ], cache_folder=tmpdir)
    p.name = ROOT

    p, outputs = p.fit_transform(
        np.array(range(10)),
        np.array(range(10))
    )

    not_saved_paths = [create_some_step3_path(tmpdir)]
    saved_paths = [create_root_path(tmpdir), create_pipeline2_path(tmpdir), create_some_step1_path(tmpdir),
                   create_some_step2_path(tmpdir), create_some_checkpoint_path(tmpdir)]
    assert np.array_equal(outputs, EXPECTED_OUTPUTS)
    for p in saved_paths:
        assert os.path.exists(p)
    for p in not_saved_paths:
        assert not os.path.exists(p)


def test_resumable_pipeline_transform_should_not_save_steps(tmpdir: LocalPath):
    p = ResumablePipeline([
        (SOME_STEP_1, MultiplyByN(multiply_by=2)),
        (PIPELINE_2, ResumablePipeline([
            (SOME_STEP_2, MultiplyByN(multiply_by=4)),
            (CHECKPOINT, DefaultCheckpoint()),
            (SOME_STEP_3, MultiplyByN(multiply_by=6)),
        ]))
    ], cache_folder=tmpdir)
    p.name = ROOT

    outputs = p.transform(
        np.array(range(10))
    )

    saved_paths = [create_root_path(tmpdir), create_pipeline2_path(tmpdir), create_some_step1_path(tmpdir),
             create_some_step2_path(tmpdir), create_some_step3_path(tmpdir), create_some_checkpoint_path(tmpdir)]
    assert np.array_equal(outputs, EXPECTED_OUTPUTS)
    for p in saved_paths:
        assert not os.path.exists(p)


def test_resumable_pipeline_fit_should_save_all_fitted_pipeline_steps(tmpdir: LocalPath):
    p = ResumablePipeline([
        (SOME_STEP_1, MultiplyByN(multiply_by=2)),
        (PIPELINE_2, ResumablePipeline([
            (SOME_STEP_2, MultiplyByN(multiply_by=4)),
            (CHECKPOINT, DefaultCheckpoint()),
            (SOME_STEP_3, MultiplyByN(multiply_by=6))
        ]))
    ], cache_folder=tmpdir)
    p.name = ROOT

    p = p.fit(
        np.array(range(10)),
        np.array(range(10))
    )

    not_saved_paths = [create_some_step3_path(tmpdir)]
    saved_paths = [create_root_path(tmpdir), create_pipeline2_path(tmpdir), create_some_step1_path(tmpdir),
                   create_some_step2_path(tmpdir), create_some_checkpoint_path(tmpdir)]
    for p in saved_paths:
        assert os.path.exists(p)
    for p in not_saved_paths:
        assert not os.path.exists(p)


def test_resumable_pipeline_fit_transform_should_load_all_pipeline_steps(tmpdir: LocalPath):
    # Given
    p = given_saved_pipeline(tmpdir)

    # When
    p, outputs = p.fit_transform(
        np.array(range(10)),
        np.array(range(10))
    )

    # Then
    assert np.array_equal(outputs, EXPECTED_OUTPUTS)


def test_resumable_pipeline_transform_should_load_all_pipeline_steps(tmpdir: LocalPath):
    # Given
    p = given_saved_pipeline(tmpdir)

    # When
    outputs = p.transform(
        np.array(range(10))
    )

    # Then
    assert np.array_equal(outputs, EXPECTED_OUTPUTS)


def test_resumable_pipeline_fit_should_load_all_pipeline_steps(tmpdir: LocalPath):
    # Given
    p = given_saved_pipeline(tmpdir)

    # When
    p = p.fit(
        np.array(range(10)),
        np.array(range(10))
    )

    # Then
    assert p[SOME_STEP_1].hyperparams['multiply_by'] == 2
    assert p[PIPELINE_2][SOME_STEP_2].hyperparams['multiply_by'] == 4
    assert p[PIPELINE_2][SOME_STEP_3].hyperparams['multiply_by'] == 6


def given_saved_pipeline(tmpdir: LocalPath):
    step_savers = [(SOME_STEP_1, []), (PIPELINE_2, [TruncableJoblibStepSaver()])]
    path = create_root_path(tmpdir, True)
    root = ResumablePipeline([], cache_folder=tmpdir)
    root.sub_steps_savers = step_savers
    root.name = ROOT
    dump(root, path)

    pipeline_2 = ResumablePipeline([], cache_folder=tmpdir)
    pipeline_2.name = 'pipeline2'
    pipeline_2.sub_steps_savers = [
        (SOME_STEP_2, []),
        (CHECKPOINT, []),
        (SOME_STEP_3, []),
    ]
    dump(pipeline_2, create_pipeline2_path(tmpdir, True))

    given_saved_some_step(multiply_by=2, name=SOME_STEP_1, path=create_some_step1_path(tmpdir, True))
    given_saved_some_step(multiply_by=4, name=SOME_STEP_2, path=create_some_step2_path(tmpdir, True))
    given_saved_some_step(multiply_by=6, name=SOME_STEP_3, path=create_some_step3_path(tmpdir, True))

    checkpoint = DefaultCheckpoint()
    checkpoint.name = CHECKPOINT
    dump(checkpoint, create_some_checkpoint_path(tmpdir, True))

    p = ResumablePipeline([
        (SOME_STEP_1, MultiplyByN(multiply_by=1)),
        (PIPELINE_2, ResumablePipeline([
            (SOME_STEP_2, MultiplyByN(multiply_by=1)),
            (CHECKPOINT, DefaultCheckpoint()),
            (SOME_STEP_3, MultiplyByN(multiply_by=1))
        ]))
    ], cache_folder=tmpdir)
    p.name = ROOT

    return p


def given_saved_some_step(multiply_by, name, path):
    some_step1 = MultiplyByN(multiply_by=multiply_by)
    some_step1.name = name
    dump(some_step1, path)



import numpy as np
import pytest
from sklearn.decomposition import PCA
from sklearn.ensemble import BaggingRegressor, GradientBoostingRegressor
from sklearn.linear_model import LinearRegression, SGDRegressor, SGDClassifier
from sklearn.metrics import median_absolute_error

from neuraxle.base import Identity
from neuraxle.hyperparams.distributions import RandInt, Uniform
from neuraxle.hyperparams.space import HyperparameterSamples
from neuraxle.hyperparams.space import HyperparameterSpace
from neuraxle.metaopt.auto_ml import KFoldCrossValidationSplitter, AutoML, RandomSearchHyperparameterSelectionStrategy, \
    HyperparamsJSONRepository
from neuraxle.metaopt.callbacks import ScoringCallback
from neuraxle.pipeline import Pipeline
from neuraxle.steps.data import DataShuffler
from neuraxle.steps.sklearn import SKLearnWrapper


def test_sklearn_wrapper_with_an_invalid_step():
    with pytest.raises(ValueError):
        SKLearnWrapper(Identity())


def test_sklearn_wrapper_fit_transform_with_predict():
    p = SKLearnWrapper(LinearRegression())
    data_inputs = np.expand_dims(np.array(list(range(10))), axis=-1)
    expected_outputs = np.expand_dims(np.array(list(range(10, 20))), axis=-1)

    p, outputs = p.fit_transform(data_inputs, expected_outputs)

    assert np.array_equal(outputs, expected_outputs)


def test_sklearn_wrapper_transform_with_predict():
    p = SKLearnWrapper(LinearRegression())
    data_inputs = np.expand_dims(np.array(list(range(10))), axis=-1)
    expected_outputs = np.expand_dims(np.array(list(range(10, 20))), axis=-1)

    p = p.fit(data_inputs, expected_outputs)
    outputs = p.transform(data_inputs)

    assert np.array_equal(outputs, expected_outputs)


def test_sklearn_wrapper_fit_transform_with_transform():
    n_components = 2
    p = SKLearnWrapper(PCA(n_components=n_components))
    dim1 = 10
    dim2 = 10
    data_inputs, expected_outputs = _create_data_source((dim1, dim2))

    p, outputs = p.fit_transform(data_inputs, expected_outputs)

    assert outputs.shape == (dim1, n_components)


def test_sklearn_wrapper_transform_partial_fit_with_predict():
    model = SKLearnWrapper(SGDRegressor(), use_partial_fit=True)
    p = Pipeline([DataShuffler(), model])
    data_inputs = np.expand_dims(np.array(list(range(10))), axis=-1)
    expected_outputs = np.ravel(np.expand_dims(np.array(list(range(10, 20))), axis=-1))

    for _ in range(2000):
        p = p.fit(data_inputs, expected_outputs)
    outputs = model.transform(data_inputs)

    assert all([np.isclose(a, b, atol=0.1) for a, b in zip(expected_outputs, outputs)])


def test_sklearn_wrapper_transform_partial_fit_classifier():
    data_inputs = np.array([[0, 1], [0, 0], [3, -2], [-1, 1], [-2, 1], [2, 0], [2, -1], [4, -2], [-3, 1], [-1, 0]])
    expected_outputs = np.ravel(np.expand_dims(data_inputs[:, 0] + 2 * data_inputs[:, 1] + 1, axis=-1))
    classes = np.array([0, 1, 2, 3])
    model = SKLearnWrapper(SGDClassifier(), use_partial_fit=True, partial_fit_kwargs={'classes': classes})
    p = Pipeline([DataShuffler(), model])

    for _ in range(2000):
        p = p.fit(data_inputs, expected_outputs)
    outputs = model.transform(data_inputs)

    assert outputs.shape == (10,)
    assert len(set(outputs) - set(classes)) == 0


def test_sklearn_wrapper_set_hyperparams():
    p = SKLearnWrapper(PCA())
    p.set_hyperparams(HyperparameterSamples({
        'n_components': 2
    }))

    assert p.wrapped_sklearn_predictor.n_components == 2


def test_sklearn_wrapper_update_hyperparams():
    p = SKLearnWrapper(PCA())
    p.set_hyperparams(HyperparameterSamples({
        'n_components': 2,
        'svd_solver': 'full'
    }))
    p.update_hyperparams(HyperparameterSamples({
        'n_components': 4
    }))

    assert p.wrapped_sklearn_predictor.n_components == 4
    assert p.wrapped_sklearn_predictor.svd_solver == 'full'


def _create_data_source(shape):
    data_inputs = np.random.random(shape).astype(np.float32)
    expected_outputs = np.random.random(shape).astype(np.float32)
    return data_inputs, expected_outputs


# With AutoML loop

def _test_within_auto_ml_loop(tmpdir, pipeline):
    X_train = np.random.random((25, 50)).astype(np.float32)
    Y_train = np.random.random((25,)).astype(np.float32)

    validation_splitter = KFoldCrossValidationSplitter(3)
    scoring_callback = ScoringCallback(
        median_absolute_error, higher_score_is_better=False)

    auto_ml = AutoML(
        pipeline=pipeline,
        hyperparams_optimizer=RandomSearchHyperparameterSelectionStrategy(),
        validation_splitter=validation_splitter,
        scoring_callback=scoring_callback,
        n_trials=2,
        epochs=1,
        hyperparams_repository=HyperparamsJSONRepository(
            cache_folder=tmpdir),
        refit_trial=True,
        continue_loop_on_error=False)

    auto_ml.fit(X_train, Y_train)


def test_automl_sklearn(tmpdir):
    grad_boost = SKLearnWrapper(GradientBoostingRegressor())
    _test_within_auto_ml_loop(tmpdir, grad_boost)


def test_automl_sklearn_model_with_base_estimator(tmpdir):
    grad_boost = GradientBoostingRegressor()
    bagged_regressor = BaggingRegressor(
        grad_boost, random_state=5, n_jobs=-1)

    wrapped_bagged_regressor = SKLearnWrapper(
        bagged_regressor,
        HyperparameterSpace({
            "n_estimators": RandInt(10, 100),
            "max_features": Uniform(0.6, 1.0)}),
        #  return_all_sklearn_default_params_on_get=True
    )
    _test_within_auto_ml_loop(tmpdir, wrapped_bagged_regressor)



import numpy as np
import pytest


from neuraxle.base import AssertExpectedOutputIsNone
from neuraxle.pipeline import Pipeline


def test_expectedoutputnull_raise_exception_when_notnull(tmpdir):
    data_inputs = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    expected_outputs = data_inputs * 2

    p = Pipeline([AssertExpectedOutputIsNone()])

    with pytest.raises(AssertionError) as error_info:
        p.fit_transform(data_inputs, expected_outputs)


def test_expectedoutputnull_is_fine_when_null(tmpdir):
    data_inputs = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    expected_outputs = None

    p = Pipeline([AssertExpectedOutputIsNone()])
    p.fit_transform(data_inputs, expected_outputs)



"""
Usage of ValueCachingWrapper in Neuraxle.
=============================================================

This demonstrates how you can use value caching in a Neuraxle pipeline.

..
    Copyright 2019, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.


"""
import os
import shutil
import time

import numpy as np

from neuraxle.pipeline import Pipeline
from neuraxle.steps.caching import JoblibValueCachingWrapper
from neuraxle.steps.loop import ForEach
from neuraxle.steps.misc import Sleep
from neuraxle.steps.numpy import MultiplyByN


def main():
    value_caching_folder = 'value_caching'
    if not os.path.exists(value_caching_folder):
        os.makedirs(value_caching_folder)

    data_inputs = list(range(100))

    sleep_time = 0.001
    a = time.time()
    for i in range(5):
        p = Pipeline([
            JoblibValueCachingWrapper(
                ForEach(Pipeline([Sleep(sleep_time=sleep_time), MultiplyByN(2)])),
                cache_folder=value_caching_folder
            )
        ])
        outputs_value_caching = p.transform(data_inputs)
    b = time.time()
    time_value_caching_pipeline = b - a
    print('Pipeline with ValueCachingWrapper')
    print('execution time: {} seconds'.format(time_value_caching_pipeline))

    a = time.time()
    for i in range(5):
        p = Pipeline([
            ForEach(Pipeline([Sleep(sleep_time=sleep_time), MultiplyByN(2)])),
        ])

        outputs_vanilla = p.transform(data_inputs)
    b = time.time()
    time_vanilla_pipeline = b - a
    print('Pipeline without value caching')
    print('execution time: {} seconds'.format(time_vanilla_pipeline))

    shutil.rmtree(value_caching_folder)

    assert np.array_equal(outputs_value_caching, outputs_vanilla)
    assert time_value_caching_pipeline < time_vanilla_pipeline


if __name__ == '__main__':
    main()



"""
Tests for pipeline hashing
========================================

..
    Copyright 2019, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

"""

from typing import List, Any

import numpy as np
import pytest

from neuraxle.base import NamedTupleList
from neuraxle.checkpoints import Checkpoint
from neuraxle.hyperparams.space import HyperparameterSamples
from neuraxle.pipeline import Pipeline, ResumablePipeline
from neuraxle.steps.misc import TransformCallbackStep, TapeCallbackFunction
from testing.test_resumable_pipeline import SomeCheckpointStep


class ResumablePipelineTestCase:
    def __init__(self, tape, steps: NamedTupleList, expected_tape: List[str], di, eo=None, expected_rehashed_ids=None):
        self.tape = tape
        self.expected_outputs = eo
        self.data_inputs = di
        self.steps = steps
        self.expected_tape = expected_tape
        self.expected_rehashed_ids = expected_rehashed_ids


def mock_hasher(current_ids, data_inputs, hyperparameters):
    if current_ids is None:
        current_ids = list(range(len(data_inputs)))
        current_ids = [str(c) for c in current_ids]
    if len(hyperparameters) == 0:
        return current_ids
    else:
        items = ",".join([str(value) for prop, value in hyperparameters.to_flat_dict().items()])
    return [
        ",".join([str(current_id), items])
        for current_id in current_ids
    ]


class ResumablePipelineWithMockHasher(ResumablePipeline):
    def hash(self, data_container):
        return mock_hasher(data_container.current_ids, data_container.data_inputs, self.hyperparams)


class TransformCallbackStepWithMockHasher(TransformCallbackStep):
    def hash(self, data_container):
        return mock_hasher(data_container.current_ids, data_container.data_inputs, self.hyperparams)


def create_callback_step(tape_step_name, hyperparams):
    step = (
        tape_step_name,
        TransformCallbackStepWithMockHasher(
            callback_function=TapeCallbackFunction().callback,
            more_arguments=[tape_step_name],
            hyperparams=HyperparameterSamples(hyperparams)
        )
    )
    return step


def find_checkpoint(steps):
    for name, step in steps:
        if isinstance(step, Pipeline):
            return find_checkpoint(step.steps_as_tuple)
        if isinstance(step, Checkpoint):
            return step
    return None


def create_test_cases():
    one_step_with_empty_hyperparameters = ResumablePipelineTestCase(
        tape=TapeCallbackFunction(),
        steps=[
            create_callback_step("a", {}),
            ("checkpoint", SomeCheckpointStep()),
        ],
        di=np.array([1, 2]),
        eo=np.array([1, 2]),
        expected_tape=["a"],
        expected_rehashed_ids=['0', '1'],
    )

    steps_with_empty_hyperparameters = ResumablePipelineTestCase(
        tape=TapeCallbackFunction(),
        steps=[
            create_callback_step("a", {}),
            create_callback_step("b", {}),
            ("checkpoint", SomeCheckpointStep()),
        ],
        di=np.array([1, 2]),
        eo=np.array([1, 2]),
        expected_tape=["a", "b"],
        expected_rehashed_ids=['0', '1'],
    )

    steps_with_empty_hyperparameters_sub_pipeline = ResumablePipelineTestCase(
        tape=TapeCallbackFunction(),
        steps=[
            create_callback_step("a", {}),
            create_callback_step("b", {}),
            Pipeline([
                create_callback_step("c", {}),
                create_callback_step("d", {}),
                ("checkpoint", SomeCheckpointStep()),
            ])
        ],
        di=np.array([1, 2]),
        eo=np.array([1, 2]),
        expected_tape=["a", "b"],
        expected_rehashed_ids=['0', '1'],
    )

    one_step_with_hyperparameters = ResumablePipelineTestCase(
        tape=TapeCallbackFunction(),
        steps=[
            create_callback_step("a", {"a__learning_rate": 1}),
            ("checkpoint", SomeCheckpointStep()),
        ],
        di=np.array([1, 2]),
        eo=np.array([1, 2]),
        expected_tape=["a"],
        expected_rehashed_ids=['0,1', '1,1'],
    )

    steps_with_hyperparameters = ResumablePipelineTestCase(
        tape=TapeCallbackFunction(),
        steps=[
            create_callback_step("a", {"a__learning_rate": 1}),
            create_callback_step("b", {"b__learning_rate": 2}),
            ("checkpoint", SomeCheckpointStep()),
        ],
        di=np.array([1, 2]),
        eo=np.array([1, 2]),
        expected_tape=["a", "b"],
        expected_rehashed_ids=['0,1,2', '1,1,2'],
    )

    steps_with_hyperparameters_sub_pipeline = ResumablePipelineTestCase(
        tape=TapeCallbackFunction(),
        steps=[
            create_callback_step("a", {"a__learning_rate": 1}),
            create_callback_step("b", {"b__learning_rate": 2}),
            Pipeline([
                create_callback_step("c", {"c__learning_rate": 3}),
                create_callback_step("d", {"d__learning_rate": 4}),
                ("checkpoint", SomeCheckpointStep()),
            ])
        ],
        di=np.array([1, 2]),
        eo=np.array([1, 2]),
        expected_tape=["a", "b"],
        expected_rehashed_ids=['0,1,2,3,4', '1,1,2,3,4'],
    )

    one_step_with_nested_hyperparameters = ResumablePipelineTestCase(
        tape=TapeCallbackFunction(),
        steps=[
            create_callback_step("a", {"a": {"learning_rate": 1}}),
            ("checkpoint", SomeCheckpointStep()),
        ],
        di=np.array([1, 2]),
        eo=np.array([1, 2]),
        expected_tape=["a"],
        expected_rehashed_ids=['0,1', '1,1'],
    )

    steps_with_nested_hyperparameters = ResumablePipelineTestCase(
        tape=TapeCallbackFunction(),
        steps=[
            create_callback_step("a", {"a": {"learning_rate": 1}}),
            create_callback_step("b", {"b": {"learning_rate": 2}}),
            ("checkpoint", SomeCheckpointStep()),
        ],
        di=np.array([1, 2]),
        eo=np.array([1, 2]),
        expected_tape=["a", "b"],
        expected_rehashed_ids=['0,1,2', '1,1,2'],
    )

    steps_with_nested_hyperparameters_sub_pipeline = ResumablePipelineTestCase(
        tape=TapeCallbackFunction(),
        steps=[
            create_callback_step("a", {"a": {"learning_rate": 1}}),
            create_callback_step("b", {"b": {"learning_rate": 2}}),
            Pipeline([
                create_callback_step("c", {"c": {"learning_rate": 3}}),
                create_callback_step("d", {"d": {"learning_rate": 4}}),
                ("checkpoint", SomeCheckpointStep()),
            ])
        ],
        di=np.array([1, 2]),
        eo=np.array([1, 2]),
        expected_tape=["a", "b"],
        expected_rehashed_ids=['0,1,2,3,4', '1,1,2,3,4'],
    )

    return [
        one_step_with_empty_hyperparameters,
        steps_with_empty_hyperparameters,
        steps_with_empty_hyperparameters_sub_pipeline,
        one_step_with_hyperparameters,
        steps_with_hyperparameters,
        steps_with_hyperparameters_sub_pipeline,
        one_step_with_nested_hyperparameters,
        steps_with_nested_hyperparameters,
        steps_with_nested_hyperparameters_sub_pipeline
    ]


@pytest.mark.parametrize("test_case", create_test_cases())
def test_transform_should_rehash_hyperparameters_for_each_steps(test_case: ResumablePipelineTestCase, tmpdir):
    pipeline = ResumablePipelineWithMockHasher(steps=test_case.steps, cache_folder=tmpdir)

    pipeline.transform(test_case.data_inputs)

    mocked_checkpoint = find_checkpoint(pipeline.steps_as_tuple)
    actual_rehashed_ids = mocked_checkpoint.saved_data_container.current_ids
    assert np.array_equal(actual_rehashed_ids, test_case.expected_rehashed_ids)


@pytest.mark.parametrize("test_case", create_test_cases())
def test_fit_should_rehash_hyperparameters_for_each_steps(test_case: ResumablePipelineTestCase, tmpdir):
    pipeline = ResumablePipelineWithMockHasher(steps=test_case.steps, cache_folder=tmpdir)

    pipeline = pipeline.fit(test_case.data_inputs, test_case.expected_outputs)

    mocked_checkpoint = find_checkpoint(pipeline.steps_as_tuple)
    actual_rehashed_ids = mocked_checkpoint.saved_data_container.current_ids
    assert np.array_equal(actual_rehashed_ids, test_case.expected_rehashed_ids)


@pytest.mark.parametrize("test_case", create_test_cases())
def test_fit_transform_should_rehash_hyperparameters_for_each_steps(test_case: ResumablePipelineTestCase, tmpdir):
    pipeline = ResumablePipelineWithMockHasher(steps=test_case.steps, cache_folder=tmpdir)

    pipeline.fit_transform(test_case.data_inputs, test_case.expected_outputs)

    mocked_checkpoint = find_checkpoint(pipeline.steps_as_tuple)
    actual_rehashed_ids = mocked_checkpoint.saved_data_container.current_ids
    assert np.array_equal(actual_rehashed_ids, test_case.expected_rehashed_ids)



"""
Usage of Checkpoints in Automatic Machine Learning (AutoML)
=============================================================

This demonstrates how you can use checkpoints in a pipeline to save computing time when doing a hyperparameter search.

..
    Copyright 2019, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

..
    Thanks to Umaneo Technologies Inc. for their contributions to this Machine Learning
    project, visit https://www.umaneo.com/ for more information on Umaneo Technologies Inc.

"""
import os
import time

import numpy as np
from sklearn.metrics import mean_squared_error

from neuraxle.checkpoints import DefaultCheckpoint
from neuraxle.hyperparams.distributions import RandInt
from neuraxle.hyperparams.space import HyperparameterSpace
from neuraxle.metaopt.auto_ml import AutoML, RandomSearchHyperparameterSelectionStrategy, ValidationSplitter
from neuraxle.metaopt.callbacks import MetricCallback, ScoringCallback
from neuraxle.pipeline import ResumablePipeline, DEFAULT_CACHE_FOLDER, Pipeline
from neuraxle.steps.flow import ExpandDim
from neuraxle.steps.loop import ForEach
from neuraxle.steps.misc import Sleep
from neuraxle.steps.numpy import MultiplyByN


def main(tmpdir, sleep_time: float = 0.001, n_iter: int = 10):
    DATA_INPUTS = np.array(range(100))
    EXPECTED_OUTPUTS = np.array(range(100, 200))

    HYPERPARAMETER_SPACE = HyperparameterSpace({
        'multiplication_1__multiply_by': RandInt(1, 2),
        'multiplication_2__multiply_by': RandInt(1, 2),
        'multiplication_3__multiply_by': RandInt(1, 2),
    })

    print('Classic Pipeline:')
    classic_pipeline_folder = os.path.join(str(tmpdir), 'classic')

    pipeline = Pipeline([
        ('multiplication_1', MultiplyByN()),
        ('sleep_1', ForEach(Sleep(sleep_time))),
        ('multiplication_2', MultiplyByN()),
        ('sleep_2', ForEach(Sleep(sleep_time))),
        ('multiplication_3', MultiplyByN()),
    ], cache_folder=classic_pipeline_folder).set_hyperparams_space(HYPERPARAMETER_SPACE)

    time_a = time.time()
    auto_ml = AutoML(
        pipeline,
        refit_trial=True,
        n_trials=n_iter,
        cache_folder_when_no_handle=classic_pipeline_folder,
        validation_splitter=ValidationSplitter(0.2),
        hyperparams_optimizer=RandomSearchHyperparameterSelectionStrategy(),
        scoring_callback=ScoringCallback(mean_squared_error, higher_score_is_better=False),
        callbacks=[
            MetricCallback('mse', metric_function=mean_squared_error, higher_score_is_better=False)
        ],
    )
    auto_ml = auto_ml.fit(DATA_INPUTS, EXPECTED_OUTPUTS)
    outputs = auto_ml.get_best_model().predict(DATA_INPUTS)
    time_b = time.time()

    actual_score = mean_squared_error(EXPECTED_OUTPUTS, outputs)
    print('{0} seconds'.format(time_b - time_a))
    print('output: {0}'.format(outputs))
    print('smallest mse: {0}'.format(actual_score))
    print('best hyperparams: {0}'.format(pipeline.get_hyperparams()))

    assert isinstance(actual_score, float)

    print('Resumable Pipeline:')
    resumable_pipeline_folder = os.path.join(str(tmpdir), 'resumable')

    pipeline = ResumablePipeline([
        ('multiplication_1', MultiplyByN()),
        ('ForEach(sleep_1)', ForEach(Sleep(sleep_time))),
        ('checkpoint1', ExpandDim(DefaultCheckpoint())),
        ('multiplication_2', MultiplyByN()),
        ('sleep_2', ForEach(Sleep(sleep_time))),
        ('checkpoint2', ExpandDim(DefaultCheckpoint())),
        ('multiplication_3', MultiplyByN())
    ], cache_folder=resumable_pipeline_folder).set_hyperparams_space(HYPERPARAMETER_SPACE)

    time_a = time.time()
    auto_ml = AutoML(
        pipeline,
        refit_trial=True,
        n_trials=n_iter,
        cache_folder_when_no_handle=resumable_pipeline_folder,
        validation_splitter=ValidationSplitter(0.2),
        hyperparams_optimizer=RandomSearchHyperparameterSelectionStrategy(),
        scoring_callback=ScoringCallback(mean_squared_error, higher_score_is_better=False),
        callbacks=[
            MetricCallback('mse', metric_function=mean_squared_error, higher_score_is_better=False)
        ]
    )
    auto_ml = auto_ml.fit(DATA_INPUTS, EXPECTED_OUTPUTS)
    outputs = auto_ml.get_best_model().predict(DATA_INPUTS)
    time_b = time.time()
    pipeline.flush_all_cache()

    actual_score = mean_squared_error(EXPECTED_OUTPUTS, outputs)
    print('{0} seconds'.format(time_b - time_a))
    print('output: {0}'.format(outputs))
    print('smallest mse: {0}'.format(actual_score))
    print('best hyperparams: {0}'.format(pipeline.get_hyperparams()))

    assert isinstance(actual_score, float)


if __name__ == "__main__":
    main(DEFAULT_CACHE_FOLDER, sleep_time=0.001, n_iter=50)



from neuraxle.pipeline import Pipeline
from neuraxle.base import BaseStep, MetaStepMixin
from neuraxle.union import Identity

from testing.mocks.step_mocks import SomeMetaStepWithHyperparams


class SomeMetaStep(MetaStepMixin, BaseStep):
    def __init__(self, wrapped: BaseStep):
        BaseStep.__init__(self)
        MetaStepMixin.__init__(self, wrapped)

    def transform(self, data_inputs):
        self.wrapped.transform(data_inputs)


def test_metastepmixin_set_train_should_set_train_to_false():
    p = SomeMetaStep(Pipeline([
        Identity()
    ]))

    p.set_train(False)

    assert not p.is_train
    assert not p.wrapped[0].is_train
    assert not p.wrapped.is_train


def test_metastepmixin_set_train_should_set_train_to_true():
    p = SomeMetaStep(Pipeline([
        Identity()
    ]))

    assert p.is_train
    assert p.wrapped[0].is_train
    assert p.wrapped.is_train


def test_basestep_str_representation_works_correctly():
    output = str(SomeMetaStepWithHyperparams())
    assert output == "SomeMetaStepWithHyperparams(SomeStepWithHyperparams(name='MockStep'), name='SomeMetaStepWithHyperparams')"



from typing import List, Tuple

from neuraxle.metaopt.auto_ml import InMemoryHyperparamsRepository, HyperparamsJSONRepository, HyperparamsRepository
from neuraxle.metaopt.observable import _Observer
from neuraxle.metaopt.trial import Trial


class SomeObserver(_Observer[Tuple[HyperparamsRepository, Trial]]):
    def __init__(self):
        self.events: List[Trial] = []

    def on_next(self, value: Tuple[HyperparamsRepository, Trial]):
        self.events.append(value)


def test_in_memory_hyperparams_repository_should_be_observable():
    repo: InMemoryHyperparamsRepository = InMemoryHyperparamsRepository()
    # todo: make a tests that asserts that an observer can receive updates from the InMemoryHyperparamsRepository
    # todo: given trial, a repo, and an observer
    pass

    # todo: when repo.subscribe(observer)
    # todo: when repo.save_trial(trial)

    # todo: then observer.events[0] == trial


def test_hyperparams_json_repository_should_be_observable_in_memory():
    # todo: make a tests that asserts that an observer can receive updates from the HyperparamsJSONRepository
    # todo: given trial, a repo, and an observer
    repo: HyperparamsJSONRepository = HyperparamsJSONRepository()

    # todo: when repo.subscribe(observer)
    # todo: when repo.save_trial(trial)

    # todo: then observer.events[0] == trial
    pass


def test_hyperparams_json_repository_should_be_observable_with_file_system_changes():
    # todo: make a tests that asserts that an observer can receive updates from the HyperparamsJSONRepository
    # todo: given trial, a repo, and an observer
    repo: HyperparamsJSONRepository = HyperparamsJSONRepository()

    # todo: when repo.subscribe_to_cache_folder_changes(observer)
    # todo: when repo.save_trial(trial)

    # todo: then observer.events[0] == trial
    pass



from typing import Any


from neuraxle.base import NamedTupleList, BaseStep, ExecutionContext
from neuraxle.pipeline import Pipeline
from testing.test_pipeline import SomeStep


class SomePipeline(Pipeline):
    def __init__(self, steps: NamedTupleList):
        Pipeline.__init__(self, steps)
        self.teared_down = False

    def teardown(self) -> 'BaseStep':
        self.teared_down = True
        return self


class SomeException(BaseStep):
    def fit_transform(self, data_inputs, expected_outputs=None) -> ('BaseStep', Any):
        raise Exception()

    def fit(self, data_inputs, expected_outputs=None) -> ('BaseStep', Any):
        raise Exception()

    def transform(self, data_inputs, expected_outputs=None) -> ('BaseStep', Any):
        raise Exception()


class SomeStepSetup(SomeStep):
    def __init__(self):
        SomeStep.__init__(self)
        self.called_with = None

    def setup(self, context: ExecutionContext = None) -> 'BaseStep':
        self.is_initialized = True
        return self


def test_fit_transform_should_setup_pipeline_and_steps():
    step_setup = SomeStepSetup()
    p = SomePipeline([
        step_setup
    ])

    assert not p.is_initialized
    assert not step_setup.is_initialized

    p.fit_transform([1], [1])

    assert p.is_initialized
    assert step_setup.is_initialized


def test_transform_should_setup_pipeline_and_steps():
    step_setup = SomeStepSetup()
    p = SomePipeline([
        step_setup
    ])
    assert not p.is_initialized
    assert not step_setup.is_initialized

    p.transform([1])

    assert p.is_initialized
    assert step_setup.is_initialized


def test_fit_should_setup_pipeline_and_steps():
    step_setup = SomeStepSetup()
    p = SomePipeline([
        step_setup
    ])
    assert not p.is_initialized
    assert not step_setup.is_initialized

    p.fit([1], [1])

    assert p.is_initialized
    assert step_setup.is_initialized


import numpy as np
from neuraxle.steps.output_handlers import InputAndOutputTransformerMixin, InputAndOutputTransformerWrapper

from neuraxle.data_container import DataContainer

from neuraxle.base import BaseTransformer, HashlibMd5ValueHasher, ExecutionContext


class WindowTimeSeries(InputAndOutputTransformerMixin, BaseTransformer):
    def __init__(self):
        BaseTransformer.__init__(self,hashers=[HashlibMd5ValueHasher()])
        InputAndOutputTransformerMixin.__init__(self)

    def transform(self, data_inputs):
        di, eo = data_inputs
        new_di, new_eo = np.array_split(np.array(di), 2)
        return new_di, new_eo


class WindowTimeSeriesForOutputTransformerWrapper(BaseTransformer):
    def __init__(self):
        BaseTransformer.__init__(self)

    def transform(self, data_inputs):
        di, eo = data_inputs
        new_di, new_eo = np.array_split(np.array(di), 2)
        return new_di, new_eo


def test_transform_input_and_output_transformer_mixin_with_hashlib_md5_value_hasher():
    data_container: DataContainer = WindowTimeSeries().handle_transform(
        data_container=DataContainer(
            data_inputs=np.array(list(range(10))),
            expected_outputs=np.array(list(range(10)))
        ),
        context=ExecutionContext()
    )

    assert np.array_equal(data_container.data_inputs, np.array(list(range(0, 5))))
    assert np.array_equal(data_container.expected_outputs, np.array(list(range(5, 10))))


def test_transform_input_and_output_transformer_wrapper_with_hashlib_md5_value_hasher():
    step = InputAndOutputTransformerWrapper(WindowTimeSeriesForOutputTransformerWrapper()) \
        .set_hashers([HashlibMd5ValueHasher()])

    data_container = step.handle_transform(
        data_container=DataContainer(
            data_inputs=np.array(list(range(10))),
            expected_outputs=np.array(list(range(10)))
        ),
        context=ExecutionContext()
    )

    assert np.array_equal(data_container.data_inputs, np.array(list(range(0, 5))))
    assert np.array_equal(data_container.expected_outputs, np.array(list(range(5, 10))))


def test_fit_transform_input_and_output_transformer_mixin_with_hashlib_md5_value_hasher():
    step, data_container = WindowTimeSeries().handle_fit_transform(
        data_container=DataContainer(
            data_inputs=np.array(list(range(10))),
            expected_outputs=np.array(list(range(10)))
        ),
        context=ExecutionContext()
    )

    assert np.array_equal(data_container.data_inputs, np.array(list(range(0, 5))))
    assert np.array_equal(data_container.expected_outputs, np.array(list(range(5, 10))))


def test_fit_transform_input_and_output_transformer_wrapper_with_hashlib_md5_value_hasher():
    step = InputAndOutputTransformerWrapper(WindowTimeSeriesForOutputTransformerWrapper()) \
        .set_hashers([HashlibMd5ValueHasher()])

    step, data_container = step.handle_fit_transform(
        data_container=DataContainer(
            data_inputs=np.array(list(range(10))),
            expected_outputs=np.array(list(range(10)))
        ),
        context=ExecutionContext()
    )

    assert np.array_equal(data_container.data_inputs, np.array(list(range(0, 5))))
    assert np.array_equal(data_container.expected_outputs, np.array(list(range(5, 10))))



import logging
import os
import shutil

import numpy as np
from sklearn.metrics import mean_squared_error

from neuraxle.base import BaseStep, ExecutionContext, HandleOnlyMixin
from neuraxle.data_container import DataContainer
from neuraxle.hyperparams.distributions import FixedHyperparameter
from neuraxle.hyperparams.space import HyperparameterSpace
from neuraxle.metaopt.auto_ml import AutoML, RandomSearchHyperparameterSelectionStrategy, ValidationSplitter, HyperparamsJSONRepository
from neuraxle.metaopt.callbacks import ScoringCallback
from neuraxle.pipeline import Pipeline
from neuraxle.steps.numpy import MultiplyByN, NumpyReshape


class LoggingStep(HandleOnlyMixin, BaseStep):
    def __init__(self):
        BaseStep.__init__(self)
        HandleOnlyMixin.__init__(self)
        self.logging_call_counter = 0

    def _fit_data_container(self, data_container: DataContainer, context: ExecutionContext) -> BaseStep:
        self._log(context, "fit")
        return self

    def _transform_data_container(self, data_container: DataContainer, context: ExecutionContext) -> DataContainer:
        self._log(context, "transform")
        return data_container

    def _fit_transform_data_container(self, data_container: DataContainer, context: ExecutionContext) -> DataContainer:
        self._log(context, "fit_transform")
        return data_container

    def _log(self, context, name):
        context.logger.warning(f"{name} call - logging call # {self.logging_call_counter}")
        self.logging_call_counter += 1


def test_logger():
    file_path = "test.log"

    if os.path.exists(file_path):
        os.remove(file_path)

    # Given
    logger = logging.getLogger('test')
    file_handler = logging.FileHandler(file_path)
    file_handler.setLevel('DEBUG')
    logger.addHandler(file_handler)
    logger.setLevel('DEBUG')
    context = ExecutionContext(logger=logger)
    pipeline = Pipeline([
        MultiplyByN(2).set_hyperparams_space(HyperparameterSpace({
            'multiply_by': FixedHyperparameter(2)
        })),
        NumpyReshape(new_shape=(-1, 1)),
        LoggingStep()
    ])

    # When
    data_container = DataContainer(
        data_inputs=np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    )
    pipeline.handle_fit(data_container, context)

    # Then
    assert os.path.exists(file_path)
    with open(file_path) as f:
        l = f.read()

    # Teardown
    file_handler.close()
    os.remove(file_path)


class TestTrialLogger:
    def test_logger_automl(self, tmpdir):
        # Given
        context = ExecutionContext()
        self.tmpdir = str(tmpdir)
        hp_repository = HyperparamsJSONRepository(cache_folder=self.tmpdir)
        n_epochs = 2
        n_trials = 4
        auto_ml = AutoML(
            pipeline=Pipeline([
                MultiplyByN(2).set_hyperparams_space(HyperparameterSpace({
                    'multiply_by': FixedHyperparameter(2)
                })),
                NumpyReshape(new_shape=(-1, 1)),
                LoggingStep()
            ]),
            hyperparams_optimizer=RandomSearchHyperparameterSelectionStrategy(),
            validation_splitter=ValidationSplitter(0.20),
            scoring_callback=ScoringCallback(mean_squared_error, higher_score_is_better=False),
            n_trials=n_trials,
            refit_trial=True,
            epochs=n_epochs,
            hyperparams_repository=hp_repository,
            continue_loop_on_error=False
        )

        # When
        data_container = DataContainer(
            data_inputs=np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]),
            expected_outputs=np.array([10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0])
        )
        auto_ml.handle_fit(data_container, context)

        # Then
        file_paths = [os.path.join(hp_repository.cache_folder, f"trial_{i}.log") for i in range(n_trials)]
        assert len(file_paths) == n_trials

        for f in file_paths:
            assert os.path.exists(f)

        for f in file_paths:
            with open(f, 'r') as f:
                log = f.readlines()
                assert len(log) == 36



import os

import joblib
from py._path.local import LocalPath

from neuraxle.checkpoints import DefaultCheckpoint
from neuraxle.pipeline import ResumablePipeline
from neuraxle.steps.misc import FitTransformCallbackStep, TapeCallbackFunction

SUMMARY_ID = '6e4419c1957e7772f3957d63bb41efcd'


def test_resumable_pipeline_with_checkpoint_fit_transform_should_save_data_inputs(tmpdir: LocalPath):
    test_case = create_checkpoint_test_case(tmpdir)

    pipeline, outputs = test_case.pipeline.fit_transform([0, 1], [1, 2])

    assert os.path.exists(os.path.join(tmpdir, 'ResumablePipeline', 'checkpoint', 'di', '0.joblib'))
    assert os.path.exists(os.path.join(tmpdir, 'ResumablePipeline', 'checkpoint', 'di', '1.joblib'))


def test_resumable_pipeline_with_checkpoint_transform_should_save_data_inputs(tmpdir):
    test_case = create_checkpoint_test_case(tmpdir)

    pipeline, outputs = test_case.pipeline.transform([0, 1])

    assert os.path.exists(os.path.join(tmpdir, 'ResumablePipeline', 'checkpoint', 'di', '0.joblib'))
    assert os.path.exists(os.path.join(tmpdir, 'ResumablePipeline', 'checkpoint', 'di', '1.joblib'))


def test_resumable_pipeline_with_checkpoint_fit_should_save_data_inputs(tmpdir):
    test_case = create_checkpoint_test_case(tmpdir)

    pipeline = test_case.pipeline.fit([0, 1], [1, 2])

    assert os.path.exists(os.path.join(tmpdir, 'ResumablePipeline', 'checkpoint', 'di', '0.joblib'))
    assert os.path.exists(os.path.join(tmpdir, 'ResumablePipeline', 'checkpoint', 'di', '1.joblib'))


def test_resumable_pipeline_with_checkpoint_fit_transform_should_save_expected_outputs(tmpdir):
    test_case = create_checkpoint_test_case(tmpdir)

    pipeline, outputs = test_case.pipeline.fit_transform([0, 1], [1, 2])

    assert os.path.exists(os.path.join(tmpdir, 'ResumablePipeline', 'checkpoint', 'eo', '0.joblib'))
    assert os.path.exists(os.path.join(tmpdir, 'ResumablePipeline', 'checkpoint', 'eo', '1.joblib'))


def test_resumable_pipeline_with_checkpoint_fit_should_save_expected_outputs(tmpdir):
    test_case = create_checkpoint_test_case(tmpdir)

    pipeline = test_case.pipeline.fit([0, 1], [1, 2])

    assert os.path.exists(os.path.join(tmpdir, 'ResumablePipeline', 'checkpoint', 'eo', '0.joblib'))
    assert os.path.exists(os.path.join(tmpdir, 'ResumablePipeline', 'checkpoint', 'eo', '1.joblib'))


def test_resumable_pipeline_with_checkpoint_fit_transform_should_resume_saved_checkpoints(tmpdir):
    given_fully_saved_checkpoints(tmpdir)
    test_case = create_checkpoint_test_case(tmpdir)

    pipeline, outputs = test_case.pipeline.fit_transform([0, 1], [1, 2])

    assert test_case.tape_transform_step1.data == [[0, 1]]
    assert test_case.tape_fit_step1.data == [([0, 1], [1, 2])]
    assert test_case.tape_fit_step2.data == [([0, 1], [1, 2])]
    assert test_case.tape_transform_step2.data == [[0, 1]]


def test_resumable_pipeline_with_checkpoint_transform_should_resume_saved_checkpoints(tmpdir):
    given_fully_saved_checkpoints(tmpdir)
    test_case = create_checkpoint_test_case(tmpdir)

    outputs = test_case.pipeline.transform([0, 1])

    assert test_case.tape_transform_step1.data == [[0, 1]]
    assert test_case.tape_fit_step2.data == []
    assert test_case.tape_transform_step2.data == [[0, 1]]


def test_resumable_pipeline_with_checkpoint_fit_should_resume_saved_checkpoints(tmpdir):
    given_fully_saved_checkpoints(tmpdir)
    test_case = create_checkpoint_test_case(tmpdir)

    pipeline = test_case.pipeline.fit([0, 1], [1, 2])

    assert test_case.tape_fit_step1.data == [([0, 1], [1, 2])]
    assert test_case.tape_transform_step1.data == [[0, 1]]
    assert test_case.tape_transform_step2.data == []
    assert test_case.tape_fit_step2.data == [([0, 1], [1, 2])]


def given_fully_saved_checkpoints(tmpdir):
    os.makedirs(os.path.join(tmpdir, 'ResumablePipeline', 'checkpoint', 'di'))
    os.makedirs(os.path.join(tmpdir, 'ResumablePipeline', 'checkpoint', 'eo'))

    with open(os.path.join(tmpdir, 'ResumablePipeline', 'checkpoint',
                           '{0}.txt'.format(SUMMARY_ID)), 'w+') as file:
        file.writelines([
            '0\n',
            '1'
        ])

    joblib.dump(0, os.path.join(tmpdir, 'ResumablePipeline', 'checkpoint', 'di', '0.joblib'))
    joblib.dump(1, os.path.join(tmpdir, 'ResumablePipeline', 'checkpoint', 'di', '1.joblib'))
    joblib.dump(1, os.path.join(tmpdir, 'ResumablePipeline', 'checkpoint', 'eo', '0.joblib'))
    joblib.dump(2, os.path.join(tmpdir, 'ResumablePipeline', 'checkpoint', 'eo', '1.joblib'))


def test_resumable_pipeline_with_checkpoint_fit_transform_should_not_resume_partially_saved_checkpoints(tmpdir):
    given_partially_saved_checkpoints(tmpdir)
    test_case = create_checkpoint_test_case(tmpdir)

    pipeline, outputs = test_case.pipeline.fit_transform([0, 1], [1, 2])

    assert test_case.tape_fit_step1.data == [([0, 1], [1, 2])]
    assert test_case.tape_fit_step2.data == [([0, 1], [1, 2])]


def test_resumable_pipeline_with_checkpoint_fit_should_not_resume_partially_saved_checkpoints(tmpdir):
    given_partially_saved_checkpoints(tmpdir)
    test_case = create_checkpoint_test_case(tmpdir)

    pipeline = test_case.pipeline.fit([0, 1], [1, 2])

    assert test_case.tape_fit_step1.data == [([0, 1], [1, 2])]
    assert test_case.tape_transform_step1.data == [[0, 1]]
    assert test_case.tape_fit_step2.data == [([0, 1], [1, 2])]
    assert test_case.tape_transform_step2.data == []


def test_resumable_pipeline_with_checkpoint_transform_should_not_resume_partially_saved_checkpoints(tmpdir):
    given_partially_saved_checkpoints(tmpdir)
    test_case = create_checkpoint_test_case(tmpdir)

    outputs = test_case.pipeline.transform([0, 1])

    assert test_case.tape_fit_step1.data == []
    assert test_case.tape_transform_step1.data == [[0, 1]]
    assert test_case.tape_fit_step2.data == []
    assert test_case.tape_transform_step2.data == [[0, 1]]


def given_partially_saved_checkpoints(tmpdir):
    os.makedirs(os.path.join(tmpdir, 'ResumablePipeline', 'checkpoint', 'di'))
    os.makedirs(os.path.join(tmpdir, 'ResumablePipeline', 'checkpoint', 'eo'))

    with open(os.path.join(tmpdir, 'ResumablePipeline', 'checkpoint',
                           '{0}.txt'.format(SUMMARY_ID)), 'w+') as file:
        file.writelines([
            '0\n',
            '1'
        ])
    joblib.dump(0, os.path.join(tmpdir, 'ResumablePipeline', 'checkpoint', 'di', '0.joblib'))
    joblib.dump(0, os.path.join(tmpdir, 'ResumablePipeline', 'checkpoint', 'eo', '1.joblib'))


class CheckpointTest:
    def __init__(self, tape_transform_step1, tape_fit_step1, tape_transform_step2, tape_fit_step2, pipeline):
        self.pipeline = pipeline
        self.tape_transform_step1 = tape_transform_step1
        self.tape_fit_step1 = tape_fit_step1

        self.tape_transform_step2 = tape_transform_step2
        self.tape_fit_step2 = tape_fit_step2


def create_checkpoint_test_case(tmpdir):
    tape_transform_1 = TapeCallbackFunction()
    tape_fit_1 = TapeCallbackFunction()
    tape_transform_2 = TapeCallbackFunction()
    tape_fit_2 = TapeCallbackFunction()
    pipeline = ResumablePipeline([
        ('step1', FitTransformCallbackStep(tape_transform_1, tape_fit_1)),
        ('checkpoint', DefaultCheckpoint()),
        ('step2', FitTransformCallbackStep(tape_transform_2, tape_fit_2))
    ], cache_folder=tmpdir)

    return CheckpointTest(
        tape_transform_1, tape_fit_1, tape_transform_2, tape_fit_2, pipeline
    )



import datetime
import time

from neuraxle.metaopt.auto_ml import InMemoryHyperparamsRepository

from neuraxle.base import Identity
from neuraxle.hyperparams.space import HyperparameterSamples
from neuraxle.metaopt.trial import Trial, TRIAL_STATUS, TRIAL_DATETIME_STR_FORMAT, Trials

EXPECTED_ERROR_TRACEBACK = 'NoneType: None\n'

EXPECTED_METRIC_RESULTS = {
    'mse': {
        'train_values': [0.5, 0.7, 0.4],
        'validation_values': [0.5, 0.7, 0.4],
        'higher_score_is_better': False
    }
}

MAIN_METRIC_NAME = 'mse'


class TestTrials:
    def setup(self):
        self.hp = HyperparameterSamples({'a': 2})
        self.repo = InMemoryHyperparamsRepository()
        self.trial = Trial(
            trial_number=0,
            save_trial_function=self.repo.save_trial,
            hyperparams=self.hp,
            main_metric_name=MAIN_METRIC_NAME
        )

    def test_trial_should_have_end_time_later_than_start_time(self):
        with self.trial.new_validation_split(Identity()) as trial_split:
            time.sleep(0.001)  # TODO: maybe remove sleep?
            trial_split.set_success()

        assert isinstance(trial_split.start_time, datetime.datetime)
        assert isinstance(trial_split.end_time, datetime.datetime)
        assert trial_split.start_time < trial_split.end_time

    def test_trial_should_create_new_split(self):
        with self.trial.new_validation_split(Identity()) as trial_split:
            trial_split.set_success()

        assert self.trial.validation_splits[0] == trial_split

    def test_trial_split_is_new_best_score_should_return_true_with_one_score(self):
        with self.trial.new_validation_split(Identity()) as trial_split:
            trial_split.add_metric_results_train(
                name=MAIN_METRIC_NAME, score=0.5, higher_score_is_better=False)
            trial_split.add_metric_results_validation(
                name=MAIN_METRIC_NAME, score=0.5, higher_score_is_better=False)

        assert trial_split.is_new_best_score()

    def test_trial_split_is_new_best_score_should_return_false_with_not_a_new_best_score(self):
        with self.trial.new_validation_split(Identity()) as trial_split:
            trial_split.add_metric_results_train(
                name=MAIN_METRIC_NAME, score=0.5, higher_score_is_better=False)
            trial_split.add_metric_results_validation(
                name=MAIN_METRIC_NAME, score=0.5, higher_score_is_better=False)

            trial_split.add_metric_results_train(
                name=MAIN_METRIC_NAME, score=0.7, higher_score_is_better=False)
            trial_split.add_metric_results_validation(
                name=MAIN_METRIC_NAME, score=0.7, higher_score_is_better=False)

        assert not trial_split.is_new_best_score()

    def test_trial_split_is_new_best_score_should_return_true_with_a_new_best_score_after_multiple_scores(self):
        with self.trial.new_validation_split(Identity()) as trial_split:
            trial_split.add_metric_results_train(
                name=MAIN_METRIC_NAME, score=0.5, higher_score_is_better=False)
            trial_split.add_metric_results_validation(
                name=MAIN_METRIC_NAME, score=0.5, higher_score_is_better=False)

            trial_split.add_metric_results_train(
                name=MAIN_METRIC_NAME, score=0.7, higher_score_is_better=False)
            trial_split.add_metric_results_validation(
                name=MAIN_METRIC_NAME, score=0.7, higher_score_is_better=False)

            trial_split.add_metric_results_train(
                name=MAIN_METRIC_NAME, score=0.4, higher_score_is_better=False)
            trial_split.add_metric_results_validation(
                name=MAIN_METRIC_NAME, score=0.4, higher_score_is_better=False)

        assert trial_split.is_new_best_score()

    def test_success_trial_split_to_json(self):
        with self.trial:
            trial_split = self._given_success_trial_validation_split(self.trial)
            trial_json = trial_split.to_json()

        self._then_success_trial_split_json_is_valid(trial_json)

    def _then_success_trial_split_json_is_valid(self, trial_json):
        assert trial_json['status'] == TRIAL_STATUS.SUCCESS.value
        assert trial_json['error'] is None
        assert trial_json['error_traceback'] is None
        assert trial_json['metric_results'] == EXPECTED_METRIC_RESULTS
        assert trial_json['main_metric_name'] == MAIN_METRIC_NAME
        start_time = datetime.datetime.strptime(trial_json['start_time'], TRIAL_DATETIME_STR_FORMAT)
        end_time = datetime.datetime.strptime(trial_json['end_time'], TRIAL_DATETIME_STR_FORMAT) + datetime.timedelta(
            hours=1)
        assert start_time < end_time

        return True

    def test_success_trial_to_json(self):
        with self.trial:
            self._given_success_trial_validation_split(self.trial)

        trial_json = self.trial.to_json()

        assert trial_json['status'] == TRIAL_STATUS.SUCCESS.value
        assert trial_json['error'] is None
        assert trial_json['error_traceback'] is None
        assert trial_json['main_metric_name'] == self.trial.main_metric_name
        assert self._then_success_trial_split_json_is_valid(trial_json['validation_splits'][0])

        start_time = datetime.datetime.strptime(
            trial_json['start_time'], TRIAL_DATETIME_STR_FORMAT)
        end_time = datetime.datetime.strptime(
            trial_json['end_time'], TRIAL_DATETIME_STR_FORMAT) + datetime.timedelta(hours=1)

        assert start_time < end_time

    def test_success_trial_get_validation_score(self):
        with self.trial:
            self._given_success_trial_validation_split(self.trial, best_score=0.3)

        validation_score = self.trial.get_validation_score()

        assert validation_score == 0.3

    def test_success_trial_multiple_splits_should_average_the_scores(self):
        with self.trial:
            self._given_success_trial_validation_split(self.trial, best_score=0.3)
            self._given_success_trial_validation_split(self.trial, best_score=0.1)

        validation_score = self.trial.get_validation_score()

        assert validation_score == 0.2

    def test_trial_with_failed_split_should_only_average_successful_splits(self):

        with self.trial:
            self._given_success_trial_validation_split(self.trial, best_score=0.3)
            self._given_success_trial_validation_split(self.trial, best_score=0.1)
            self._given_failed_trial_split(self.trial)

        validation_score = self.trial.get_validation_score()

        assert validation_score == 0.2

    def _given_success_trial_validation_split(self, trial, best_score=0.4):
        with trial.new_validation_split(Identity()) as trial_split:
            trial_split.add_metric_results_train(
                name=MAIN_METRIC_NAME, score=0.5, higher_score_is_better=False)
            trial_split.add_metric_results_validation(
                name=MAIN_METRIC_NAME, score=0.5,
                higher_score_is_better=False)

            trial_split.add_metric_results_train(
                name=MAIN_METRIC_NAME, score=0.7, higher_score_is_better=False)
            trial_split.add_metric_results_validation(
                name=MAIN_METRIC_NAME, score=0.7, higher_score_is_better=False)

            trial_split.add_metric_results_train(
                name=MAIN_METRIC_NAME, score=best_score, higher_score_is_better=False)
            trial_split.add_metric_results_validation(
                name=MAIN_METRIC_NAME, score=best_score, higher_score_is_better=False)

            trial_split.set_success()
            trial.set_success()

        return trial_split

    def test_failure_trial_split_to_json(self):
        with self.trial:
            trial_split = self._given_failed_trial_split(self.trial)

        trial_json = trial_split.to_json()

        self._then_failed_validation_split_json_is_valid(trial_json, trial_split)

    def _then_failed_validation_split_json_is_valid(self, trial_json, trial_split):
        assert trial_json['status'] == TRIAL_STATUS.FAILED.value
        assert trial_json['error'] == str(trial_split.error)
        assert trial_json['error_traceback'] == EXPECTED_ERROR_TRACEBACK
        assert trial_json['metric_results'] == EXPECTED_METRIC_RESULTS
        assert trial_json['main_metric_name'] == trial_split.main_metric_name

        start_time = datetime.datetime.strptime(
            trial_json['start_time'], TRIAL_DATETIME_STR_FORMAT)
        end_time = datetime.datetime.strptime(
            trial_json['end_time'], TRIAL_DATETIME_STR_FORMAT) + datetime.timedelta(hours=1)
        assert start_time < end_time
        return True

    def test_failure_trial_to_json(self):
        with self.trial:
            trial_split = self._given_failed_trial_split(self.trial)

        trial_json = self.trial.to_json()

        assert trial_json['status'] == TRIAL_STATUS.FAILED.value
        assert trial_json['error'] == str(trial_split.error)
        assert trial_json['error_traceback'] == EXPECTED_ERROR_TRACEBACK
        assert trial_json['main_metric_name'] == self.trial.main_metric_name
        assert self._then_failed_validation_split_json_is_valid(
            trial_json['validation_splits'][0], trial_split=trial_split)

        start_time = datetime.datetime.strptime(
            trial_json['start_time'], TRIAL_DATETIME_STR_FORMAT)
        end_time = datetime.datetime.strptime(
            trial_json['end_time'], TRIAL_DATETIME_STR_FORMAT) + datetime.timedelta(hours=1)

        assert start_time < end_time

    def _given_failed_trial_split(self, trial):
        with trial.new_validation_split(Identity()) as trial_split:
            trial_split.add_metric_results_train(
                name=MAIN_METRIC_NAME, score=0.5, higher_score_is_better=False)
            trial_split.add_metric_results_validation(
                name=MAIN_METRIC_NAME, score=0.5, higher_score_is_better=False)

            trial_split.add_metric_results_train(
                name=MAIN_METRIC_NAME, score=0.7, higher_score_is_better=False)
            trial_split.add_metric_results_validation(
                name=MAIN_METRIC_NAME, score=0.7, higher_score_is_better=False)

            trial_split.add_metric_results_train(
                name=MAIN_METRIC_NAME, score=0.4, higher_score_is_better=False)
            trial_split.add_metric_results_validation(
                name=MAIN_METRIC_NAME, score=0.4, higher_score_is_better=False)
            error = IndexError('index error')
            trial_split.set_failed(error)
            trial.set_failed(error)
        return trial_split

    def test_trials_get_best_hyperparams_should_return_hyperparams_of_best_trial(self):
        # Given
        trial_1 = self.trial
        with trial_1:
            self._given_success_trial_validation_split(trial_1, best_score=0.2)

        hp_trial_2 = HyperparameterSamples({'b': 3})
        trial_2 = Trial(
            trial_number=1, save_trial_function=self.repo.save_trial,
            hyperparams=hp_trial_2, main_metric_name=MAIN_METRIC_NAME)
        with trial_2:
            self._given_success_trial_validation_split(trial_2, best_score=0.1)

        trials = Trials(trials=[trial_1, trial_2])

        # When
        best_hyperparams = trials.get_best_hyperparams()

        # Then
        assert best_hyperparams == hp_trial_2



import numpy as np
import pytest
from sklearn import linear_model
from sklearn.exceptions import ConvergenceWarning
from sklearn.metrics import mean_squared_error
from sklearn.svm import LinearSVC
from sklearn.utils._testing import ignore_warnings

from neuraxle.base import ExecutionContext
from neuraxle.data_container import DataContainer
from neuraxle.hyperparams.distributions import FixedHyperparameter, RandInt
from neuraxle.hyperparams.space import HyperparameterSpace
from neuraxle.metaopt.auto_ml import InMemoryHyperparamsRepository, AutoML, RandomSearchHyperparameterSelectionStrategy, \
    HyperparamsJSONRepository, \
    ValidationSplitter, KFoldCrossValidationSplitter, Trainer
from neuraxle.metaopt.callbacks import MetricCallback, ScoringCallback, EarlyStoppingCallback, BestModelCheckpoint
from neuraxle.metaopt.trial import Trial, Trials
from neuraxle.pipeline import Pipeline
from neuraxle.steps.numpy import MultiplyByN, NumpyReshape
from neuraxle.steps.sklearn import SKLearnWrapper


def test_automl_early_stopping_callback(tmpdir):
    # Given
    hp_repository = InMemoryHyperparamsRepository(cache_folder=str(tmpdir))
    n_epochs = 10
    max_epochs_without_improvement = 3
    auto_ml = AutoML(
        pipeline=Pipeline([
            MultiplyByN(2).set_hyperparams_space(HyperparameterSpace({
                'multiply_by': FixedHyperparameter(2)
            })),
            NumpyReshape(new_shape=(-1, 1)),
        ]),
        hyperparams_optimizer=RandomSearchHyperparameterSelectionStrategy(),
        validation_splitter=ValidationSplitter(0.20),
        scoring_callback=ScoringCallback(mean_squared_error, higher_score_is_better=False),
        callbacks=[
            MetricCallback('mse', metric_function=mean_squared_error, higher_score_is_better=False),
            EarlyStoppingCallback(max_epochs_without_improvement)
        ],
        n_trials=1,
        refit_trial=True,
        epochs=n_epochs,
        hyperparams_repository=hp_repository,
        continue_loop_on_error=False
    )

    # When
    data_inputs = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    expected_outputs = data_inputs * 2
    auto_ml.fit(data_inputs=data_inputs, expected_outputs=expected_outputs)

    # Then
    trial = hp_repository.trials[0]
    assert len(trial.validation_splits) == 1
    validation_scores = trial.validation_splits[0].get_validation_scores()
    nepochs_executed = len(validation_scores)
    assert nepochs_executed == max_epochs_without_improvement + 1


def test_automl_savebestmodel_callback(tmpdir):
    # Given
    hp_repository = HyperparamsJSONRepository(cache_folder=tmpdir)
    validation_splitter = ValidationSplitter(0.20)
    auto_ml = AutoML(
        pipeline=Pipeline([
            MultiplyByN(2).set_hyperparams_space(HyperparameterSpace({
                'multiply_by': FixedHyperparameter(2)
            })),
            NumpyReshape(new_shape=(-1, 1)),
            linear_model.LinearRegression()
        ]),
        validation_splitter=validation_splitter,
        hyperparams_optimizer=RandomSearchHyperparameterSelectionStrategy(),
        scoring_callback=ScoringCallback(mean_squared_error, higher_score_is_better=False),
        callbacks=[
            BestModelCheckpoint()
        ],
        n_trials=3,
        epochs=1,
        refit_trial=False,
        hyperparams_repository=hp_repository,
        continue_loop_on_error=False
    ).with_context(ExecutionContext(tmpdir))

    data_inputs = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    expected_outputs = data_inputs * 4

    # When
    auto_ml.fit(data_inputs=data_inputs, expected_outputs=expected_outputs)

    # Then
    trials: Trials = hp_repository.load_all_trials()
    best_trial = trials.get_best_trial()
    best_trial_score = best_trial.get_validation_score()
    best_model = best_trial.get_model('best')
    _, _, _, valid_inputs, valid_outputs, _ = validation_splitter.split(data_inputs=data_inputs, expected_outputs=expected_outputs)
    predicted_output = best_model.predict(*valid_inputs)
    score = mean_squared_error(*valid_outputs, predicted_output)

    assert best_trial_score == score


def test_automl_with_kfold(tmpdir):
    # Given
    hp_repository = HyperparamsJSONRepository(cache_folder=tmpdir)
    auto_ml = AutoML(
        pipeline=Pipeline([
            MultiplyByN(2).set_hyperparams_space(HyperparameterSpace({
                'multiply_by': FixedHyperparameter(2)
            })),
            NumpyReshape(new_shape=(-1, 1)),
            linear_model.LinearRegression()
        ]),
        validation_splitter=ValidationSplitter(0.20),
        hyperparams_optimizer=RandomSearchHyperparameterSelectionStrategy(),
        scoring_callback=ScoringCallback(mean_squared_error, higher_score_is_better=False),
        callbacks=[
            MetricCallback('mse', metric_function=mean_squared_error,
                           higher_score_is_better=False),
        ],
        n_trials=1,
        epochs=10,
        refit_trial=True,
        hyperparams_repository=hp_repository,
        continue_loop_on_error=False
    )

    data_inputs = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    expected_outputs = data_inputs * 4

    # When
    auto_ml.fit(data_inputs=data_inputs, expected_outputs=expected_outputs)

    # Then
    p = auto_ml.get_best_model()
    outputs = p.transform(data_inputs)
    mse = mean_squared_error(expected_outputs, outputs)

    assert mse < 1000


def test_validation_splitter_should_split_data_properly():
    # Given
    data_inputs = np.random.random((4, 2, 2048, 6)).astype(np.float32)
    expected_outputs = np.random.random((4, 2, 2048, 1)).astype(np.float32)
    splitter = ValidationSplitter(test_size=0.2)

    # When
    validation_splits = splitter.split_data_container(
        data_container=DataContainer(data_inputs=data_inputs, expected_outputs=expected_outputs),
        context=ExecutionContext()
    )
    train_di, train_eo, validation_di, validation_eo = extract_validation_split_data(validation_splits)

    train_di = train_di[0]
    train_eo = train_eo[0]

    validation_di = validation_di[0]
    validation_eo = validation_eo[0]

    # Then
    assert len(train_di) == 3
    assert np.array_equal(np.array(train_di), data_inputs[0:3])
    assert len(train_eo) == 3
    assert np.array_equal(np.array(train_eo), expected_outputs[0:3])

    assert len(validation_di) == 1
    assert np.array_equal(validation_di[0], data_inputs[-1])
    assert len(validation_eo) == 1
    assert np.array_equal(validation_eo[0], expected_outputs[-1])


def test_kfold_cross_validation_should_split_data_properly():
    # Given
    data_inputs = np.random.random((4, 2, 2048, 6)).astype(np.float32)
    expected_outputs = np.random.random((4, 2, 2048, 1)).astype(np.float32)
    splitter = KFoldCrossValidationSplitter(k_fold=4)

    # When
    validation_splits = splitter.split_data_container(
        data_container=DataContainer(data_inputs=data_inputs, expected_outputs=expected_outputs),
        context=ExecutionContext()
    )
    train_di, train_eo, validation_di, validation_eo = extract_validation_split_data(validation_splits)

    # Then
    assert len(train_di[0]) == 3
    assert np.array_equal(np.array(train_di[0]), data_inputs[1:])
    assert len(train_eo[0]) == 3
    assert np.array_equal(np.array(train_eo[0]), expected_outputs[1:])

    assert len(train_di[1]) == 3
    assert np.array_equal(np.array(train_di[1]),
                          np.concatenate((np.expand_dims(data_inputs[0], axis=0), data_inputs[2:]), axis=0))
    assert len(train_eo[1]) == 3
    assert np.array_equal(np.array(train_eo[1]),
                          np.concatenate((np.expand_dims(expected_outputs[0], axis=0), expected_outputs[2:]), axis=0))

    assert len(train_di[2]) == 3
    assert np.array_equal(np.array(train_di[2]),
                          np.concatenate((data_inputs[0:2], np.expand_dims(data_inputs[3], axis=0)), axis=0))
    assert len(train_eo[2]) == 3
    assert np.array_equal(np.array(train_eo[2]),
                          np.concatenate((expected_outputs[0:2], np.expand_dims(expected_outputs[3], axis=0)), axis=0))

    assert len(train_di[3]) == 3
    assert np.array_equal(np.array(train_di[3]), data_inputs[0:3])
    assert len(train_eo[3]) == 3
    assert np.array_equal(np.array(train_eo[3]), expected_outputs[0:3])

    assert len(validation_di[0]) == 1
    assert np.array_equal(validation_di[0][0], data_inputs[0])
    assert len(validation_eo[0]) == 1
    assert np.array_equal(validation_eo[0][0], expected_outputs[0])

    assert len(validation_di[1]) == 1
    assert np.array_equal(validation_di[1][0], data_inputs[1])
    assert len(validation_eo[1]) == 1
    assert np.array_equal(validation_eo[1][0], expected_outputs[1])

    assert len(validation_di[2]) == 1
    assert np.array_equal(validation_di[2][0], data_inputs[2])
    assert len(validation_eo[2]) == 1
    assert np.array_equal(validation_eo[2][0], expected_outputs[2])

    assert len(validation_di[3]) == 1
    assert np.array_equal(validation_di[3][0], data_inputs[3])
    assert len(validation_eo[3]) == 1
    assert np.array_equal(validation_eo[3][0], expected_outputs[3])


def test_kfold_cross_validation_should_split_data_properly_bug():
    data_inputs = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    expected_outputs = np.array([0, 4, 8, 12, 16, 20, 24, 28, 32, 36, 40])
    data_container = DataContainer(
        data_inputs=data_inputs,
        expected_outputs=expected_outputs
    )
    splitter = KFoldCrossValidationSplitter(k_fold=2)

    # When
    validation_splits = splitter.split_data_container(data_container, ExecutionContext())

    train_di, train_eo, validation_di, validation_eo = extract_validation_split_data(validation_splits)

    # Then
    assert len(train_di[0]) == 6
    assert np.array_equal(np.array(train_di[0]), data_inputs[5:])
    assert len(train_eo[0]) == 6
    assert np.array_equal(np.array(train_eo[0]), expected_outputs[5:])

    assert len(train_di[1]) == 5
    assert np.array_equal(
        np.array(train_di[1]),
        data_inputs[:5]
    )
    assert len(train_eo[1]) == 5
    assert np.array_equal(
        np.array(train_eo[1]),
        expected_outputs[:5]
    )

    assert len(validation_di[0]) == 5
    assert np.array_equal(np.array(validation_di[0]), data_inputs[:5])
    assert len(validation_eo[0]) == 5
    assert np.array_equal(np.array(validation_eo[0]), expected_outputs[:5])

    assert len(validation_di[1]) == 6
    assert np.array_equal(np.array(validation_di[1]), data_inputs[5:])
    assert len(validation_eo[1]) == 6
    assert np.array_equal(validation_eo[1], expected_outputs[5:])


def extract_validation_split_data(validation_splits):
    train_di = []
    train_eo = []
    validation_di = []
    validation_eo = []
    for train_dc, validation_dc in validation_splits:
        train_di.append(train_dc.data_inputs)
        train_eo.append(train_dc.expected_outputs)

        validation_di.append(validation_dc.data_inputs)
        validation_eo.append(validation_dc.expected_outputs)
    return train_di, train_eo, validation_di, validation_eo


@ignore_warnings(category=ConvergenceWarning)
def test_automl_should_shallow_copy_data_before_each_epoch():
    # see issue #332 https://github.com/Neuraxio/Neuraxle/issues/332
    data_inputs = np.random.randint(0, 100, (100, 3))
    expected_outputs = np.array(np.sum(data_inputs, axis=-1) / 100, dtype=int)
    data_inputs += np.random.randint(0, 100, (100, 3))

    from sklearn.preprocessing import StandardScaler
    p = Pipeline([
        SKLearnWrapper(StandardScaler()),
        SKLearnWrapper(LinearSVC(), HyperparameterSpace({'C': RandInt(0, 10000)})),
    ])

    auto_ml = AutoML(
        p,
        validation_splitter=ValidationSplitter(0.20),
        refit_trial=True,
        n_trials=10,
        epochs=1,
        cache_folder_when_no_handle='cache',
        scoring_callback=ScoringCallback(mean_squared_error, higher_score_is_better=False),
        callbacks=[MetricCallback('mse', metric_function=mean_squared_error, higher_score_is_better=False)],
        hyperparams_repository=InMemoryHyperparamsRepository(
            cache_folder='cache'),
        continue_loop_on_error=False
    )

    random_search = auto_ml.fit(data_inputs, expected_outputs)

    best_model = random_search.get_best_model()

    assert isinstance(best_model, Pipeline)


def test_trainer_train():
    data_inputs = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    expected_outputs = data_inputs * 4
    p = Pipeline([
        MultiplyByN(2).set_hyperparams_space(HyperparameterSpace({
            'multiply_by': FixedHyperparameter(2)
        })),
        NumpyReshape(new_shape=(-1, 1)),
        linear_model.LinearRegression()
    ])

    trainer: Trainer = Trainer(
        epochs=10,
        scoring_callback=ScoringCallback(mean_squared_error, higher_score_is_better=False),
        validation_splitter=ValidationSplitter(test_size=0.20)
    )

    repo_trial: Trial = trainer.train(pipeline=p, data_inputs=data_inputs,
                                      expected_outputs=expected_outputs, context=ExecutionContext())

    trained_pipeline = repo_trial.get_trained_pipeline(split_number=0)

    outputs = trained_pipeline.transform(data_inputs)
    mse = mean_squared_error(expected_outputs, outputs)

    assert mse < 1



import json
import os

from neuraxle.hyperparams.space import HyperparameterSamples
from neuraxle.metaopt.deprecated import HyperparamsJSONRepository

HYPERPARAMS = {'learning_rate': 0.01}
FIRST_SCORE_FOR_TRIAL = 1


def test_hyperparams_repository_should_create_new_trial(tmpdir):
    hyperparams_json_repository = HyperparamsJSONRepository(tmpdir)
    hyperparams = HyperparameterSamples(HYPERPARAMS)

    hyperparams_json_repository.create_new_trial(hyperparams)

    trial_hash = hyperparams_json_repository._get_trial_hash(hyperparams.to_flat_dict())
    file_name = 'NEW_' + trial_hash + '.json'
    path = os.path.join(tmpdir, file_name)
    with open(path) as f:
        trial_json = json.load(f)
    assert trial_json['hyperparams'] == hyperparams.to_flat_dict()
    assert trial_json['score'] is None


def test_hyperparams_repository_should_load_all_trials(tmpdir):
    tmpdir = os.path.join(tmpdir, "__json__")
    os.mkdir(tmpdir)
    hyperparams_json_repository = HyperparamsJSONRepository(tmpdir)
    n_trials = 3
    for i in range(n_trials):
        hyperparams = HyperparameterSamples({'learning_rate': 0.01 + i * 0.01})
        hyperparams_json_repository.save_score_for_success_trial(hyperparams, i)

    trials = hyperparams_json_repository.load_all_trials()

    assert len(trials) == n_trials
    for i in range(n_trials):
        assert trials[i].hyperparams == HyperparameterSamples(
            {'learning_rate': 0.01 + i * 0.01}).to_flat_dict(), (i, str(trials))


def test_hyperparams_repository_should_save_failed_trial(tmpdir):
    hyperparams_json_repository = HyperparamsJSONRepository(tmpdir)
    hyperparams = HyperparameterSamples(HYPERPARAMS)

    hyperparams_json_repository._save_failed_trial_json(hyperparams, Exception('trial failed'))

    trial_hash = hyperparams_json_repository._get_trial_hash(hyperparams.to_flat_dict())
    file_name = 'FAILED_' + trial_hash + '.json'
    path = os.path.join(tmpdir, file_name)
    with open(path) as f:
        trial_json = json.load(f)
    assert trial_json['hyperparams'] == hyperparams.to_flat_dict()
    assert 'exception' in trial_json.keys()
    assert trial_json['score'] is None


def test_hyperparams_repository_should_save_success_trial(tmpdir):
    hyperparams_json_repository = HyperparamsJSONRepository(tmpdir)
    hyperparams = HyperparameterSamples(HYPERPARAMS)

    hyperparams_json_repository._save_successful_trial_json(hyperparams, FIRST_SCORE_FOR_TRIAL)

    trial_hash = hyperparams_json_repository._get_trial_hash(hyperparams.to_flat_dict())
    file_name = str(float(FIRST_SCORE_FOR_TRIAL)).replace('.', ',') + '_' + trial_hash + '.json'
    path = os.path.join(tmpdir, file_name)
    with open(path) as f:
        trial_json = json.load(f)
    assert trial_json['hyperparams'] == hyperparams.to_flat_dict()
    assert trial_json['score'] == FIRST_SCORE_FOR_TRIAL



"""
Tests for joblib checkpoint steps
========================================

..
    Copyright 2019, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

"""

import os

import joblib
import numpy as np
from py._path.local import LocalPath

from neuraxle.base import ExecutionContext, Identity
from neuraxle.checkpoints import DefaultCheckpoint
from neuraxle.data_container import DataContainer
from neuraxle.hyperparams.space import HyperparameterSamples
from neuraxle.pipeline import ResumablePipeline
from neuraxle.steps.misc import TapeCallbackFunction, TransformCallbackStep, BaseCallbackStep
from testing.steps.test_output_transformer_wrapper import MultiplyBy2OutputTransformer

EXPECTED_TAPE_AFTER_CHECKPOINT = ["2", "3"]

data_inputs = np.array([1, 2])
expected_outputs = np.array([2, 3])
expected_rehashed_data_inputs = ['44f9d6dd8b6ccae571ca04525c3eaffa', '898a67b2f5eeae6393ca4b3162ba8e3d']


class DifferentCallbackStep(BaseCallbackStep):
    def transform(self, data_inputs):
        self._callback(data_inputs)
        return data_inputs


def create_pipeline(tmpdir, joblib_checkpoint_step, tape, hyperparameters=None, different=False, save_pipeline=True):
    if different:
        pipeline = ResumablePipeline(
            steps=[
                ('a',
                 DifferentCallbackStep(tape.callback, ["1"], hyperparams=hyperparameters)),
                ('joblib_checkpoint', joblib_checkpoint_step),
                ('c', TransformCallbackStep(tape.callback, ["2"])),
                ('d', TransformCallbackStep(tape.callback, ["3"]))
            ],
            cache_folder=tmpdir
        )
    else:
        pipeline = ResumablePipeline(
            steps=[
                ('a',
                 TransformCallbackStep(tape.callback, ["1"], hyperparams=hyperparameters)),
                ('joblib_checkpoint', joblib_checkpoint_step),
                ('c', TransformCallbackStep(tape.callback, ["2"])),
                ('d', TransformCallbackStep(tape.callback, ["3"]))
            ], cache_folder=tmpdir
        )
    return pipeline


def test_when_no_hyperparams_should_save_checkpoint_joblib(tmpdir: LocalPath):
    tape = TapeCallbackFunction()
    joblib_checkpoint_step = DefaultCheckpoint()
    pipeline = create_pipeline(tmpdir, joblib_checkpoint_step, tape)

    pipeline, actual_data_inputs = pipeline.fit_transform(data_inputs, expected_outputs)

    actual_tape = tape.get_name_tape()
    assert np.array_equal(actual_data_inputs, data_inputs)
    assert actual_tape == ["1", "2", "3"]
    assert os.path.exists(os.path.join(tmpdir, 'ResumablePipeline', 'joblib_checkpoint', 'di', '0.joblib'))
    assert os.path.exists(os.path.join(tmpdir, 'ResumablePipeline', 'joblib_checkpoint', 'di', '1.joblib'))
    assert os.path.exists(os.path.join(tmpdir, 'ResumablePipeline', 'joblib_checkpoint', 'eo', '0.joblib'))
    assert os.path.exists(os.path.join(tmpdir, 'ResumablePipeline', 'joblib_checkpoint', 'eo', '1.joblib'))


def test_when_hyperparams_should_save_checkpoint_joblib(tmpdir: LocalPath):
    tape = TapeCallbackFunction()
    joblib_checkpoint_step = DefaultCheckpoint()
    pipeline = create_pipeline(tmpdir, joblib_checkpoint_step, tape,
                               HyperparameterSamples({"a__learning_rate": 1}))

    pipeline, actual_data_inputs = pipeline.fit_transform(data_inputs, expected_outputs)

    actual_tape = tape.get_name_tape()
    assert np.array_equal(actual_data_inputs, data_inputs)
    assert actual_tape == ["1", "2", "3"]

    assert os.path.exists(
        os.path.join(tmpdir, 'ResumablePipeline', 'joblib_checkpoint', 'di', '44f9d6dd8b6ccae571ca04525c3eaffa.joblib'))
    assert os.path.exists(
        os.path.join(tmpdir, 'ResumablePipeline', 'joblib_checkpoint', 'di', '898a67b2f5eeae6393ca4b3162ba8e3d.joblib'))
    assert os.path.exists(
        os.path.join(tmpdir, 'ResumablePipeline', 'joblib_checkpoint', 'eo', '44f9d6dd8b6ccae571ca04525c3eaffa.joblib'))
    assert os.path.exists(
        os.path.join(tmpdir, 'ResumablePipeline', 'joblib_checkpoint', 'eo', '898a67b2f5eeae6393ca4b3162ba8e3d.joblib'))


def test_when_no_hyperparams_and_saved_same_pipeline_should_load_checkpoint_joblib(tmpdir: LocalPath):
    # Given
    tape = TapeCallbackFunction()

    # When
    pipeline_save = create_pipeline(
        tmpdir=tmpdir,
        joblib_checkpoint_step=DefaultCheckpoint(),
        tape=TapeCallbackFunction()
    )
    pipeline_save.fit_transform(data_inputs, expected_outputs)

    pipeline_load = create_pipeline(
        tmpdir=tmpdir,
        joblib_checkpoint_step=DefaultCheckpoint(),
        tape=tape
    )
    pipeline_load, actual_data_inputs = pipeline_load.fit_transform(data_inputs, expected_outputs)

    # Then
    actual_tape = tape.get_name_tape()
    assert np.array_equal(actual_data_inputs, data_inputs)
    assert actual_tape == EXPECTED_TAPE_AFTER_CHECKPOINT


def test_when_hyperparams_and_saved_same_pipeline_should_load_checkpoint_joblib(tmpdir: LocalPath):
    # Given
    tape = TapeCallbackFunction()

    # When
    pipeline_save = create_pipeline(
        tmpdir=tmpdir,
        joblib_checkpoint_step=DefaultCheckpoint(),
        tape=TapeCallbackFunction(),
        hyperparameters=HyperparameterSamples({"a__learning_rate": 1})
    )
    pipeline_save.fit_transform(data_inputs, expected_outputs)

    pipeline_load = create_pipeline(
        tmpdir=tmpdir,
        joblib_checkpoint_step=DefaultCheckpoint(),
        tape=tape,
        hyperparameters=HyperparameterSamples({"a__learning_rate": 1})
    )
    pipeline_load, actual_data_inputs = pipeline_load.fit_transform(data_inputs, expected_outputs)

    # Then
    actual_tape = tape.get_name_tape()
    assert np.array_equal(actual_data_inputs, data_inputs)
    assert actual_tape == EXPECTED_TAPE_AFTER_CHECKPOINT


def test_when_hyperparams_and_saved_no_pipeline_should_not_load_checkpoint_joblib(tmpdir: LocalPath):
    # Given
    tape = TapeCallbackFunction()
    joblib_checkpoint_step = DefaultCheckpoint()

    # When
    pipeline_save = create_pipeline(
        tmpdir=tmpdir,
        joblib_checkpoint_step=Identity(),
        tape=TapeCallbackFunction(),
        hyperparameters=HyperparameterSamples({"a__learning_rate": 1}),
        different=True,
        save_pipeline=False
    )
    pipeline_save.fit_transform(data_inputs, expected_outputs)

    pipeline_load = create_pipeline(
        tmpdir=tmpdir,
        joblib_checkpoint_step=joblib_checkpoint_step,
        tape=tape,
        hyperparameters=HyperparameterSamples({"a__learning_rate": 1})
    )
    pipeline_load, actual_data_inputs = pipeline_load.fit_transform(data_inputs, expected_outputs)

    # Then
    actual_tape = tape.get_name_tape()
    assert np.array_equal(actual_data_inputs, data_inputs)
    assert actual_tape == ["1", "2", "3"]


def setup_joblib_checkpoint(current_id, data_input, expected_output, joblib_checkpoint_step):
    file = joblib_checkpoint_step.get_checkpoint_filename_path_for_current_id(current_id)
    joblib.dump((current_id, data_input, expected_output), file)


def create_pipeline_output_transformer(tmpdir):
    return ResumablePipeline([
        ('output_transformer_1', MultiplyBy2OutputTransformer()),
        ('joblib_checkpoint', DefaultCheckpoint()),
        ('output_transformer_2', MultiplyBy2OutputTransformer()),
    ], cache_folder=tmpdir)


def test_joblib_checkpoint_step_should_load_data_container(tmpdir: LocalPath):
    initial_data_inputs = [1, 2]
    initial_expected_outputs = [2, 3]

    create_pipeline_output_transformer(tmpdir).fit_transform(
        data_inputs=initial_data_inputs, expected_outputs=initial_expected_outputs
    )

    actual_data_container = create_pipeline_output_transformer(tmpdir).handle_transform(
        DataContainer(data_inputs=initial_data_inputs, current_ids=[0, 1], expected_outputs=initial_expected_outputs),
        ExecutionContext(tmpdir)
    )

    assert np.array_equal(actual_data_container.data_inputs, [4, 8])
    assert np.array_equal(actual_data_container.expected_outputs, [8, 12])
